"""Server lifecycle commands for the Bloom CLI."""

from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from cli_core_yo.registry import CommandRegistry
    from cli_core_yo.spec import CliSpec

import json
import os
import subprocess
import sys
import time
from pathlib import Path

import typer
from cli_core_yo.certs import resolve_https_certs, shared_dayhoff_certs_dir
from cli_core_yo.server import (
    display_host,
    latest_log,
    new_log_path,
    read_pid,
    stop_pid,
    write_pid,
)
from rich.console import Console

from bloom_lims.config import (
    DEFAULT_BLOOM_WEB_PORT,
    apply_runtime_environment,
    atlas_webhook_secret_warning,
    get_settings,
    get_tapdb_db_config,
)

server_app = typer.Typer(help="Server lifecycle commands.", no_args_is_help=True)
console = Console()

PROJECT_ROOT = Path(__file__).resolve().parents[2]
TAPDB_LOG_DIR = Path.home() / ".config" / "tapdb" / "logs"
SERVER_META_FILE = "server-meta.json"


class LogService(str, Enum):
    server = "server"
    tapdb = "tapdb"
    all = "all"


def _state_dir() -> Path:
    """Return the Bloom XDG state directory."""
    try:
        from cli_core_yo.runtime import get_context

        return get_context().xdg_paths.state
    except Exception:
        return Path.home() / ".local" / "state" / "bloom"


def _log_dir() -> Path:
    return _state_dir() / "logs"


def _pid_file() -> Path:
    return _state_dir() / "server.pid"


def _runtime_meta_file() -> Path:
    return _state_dir() / SERVER_META_FILE


def _ensure_dir() -> None:
    _state_dir().mkdir(parents=True, exist_ok=True)
    _log_dir().mkdir(parents=True, exist_ok=True)


def _latest_server_log() -> Path | None:
    return latest_log(_log_dir())


def active_server_pid() -> tuple[int | None, Path]:
    pid_file = _pid_file()
    return read_pid(pid_file), pid_file


def server_status_label() -> str:
    pid, _ = active_server_pid()
    if pid is None:
        return "Stopped"
    return f"Running ({_runtime_scheme().upper()}, PID {pid})"


def _runtime_host_and_port(default_port: int, default_host: str) -> tuple[str, int]:
    host = os.environ.get(
        "BLOOM_RUNTIME__HOST", os.environ.get("BLOOM_HOST", default_host)
    )
    port = int(
        os.environ.get(
            "BLOOM_RUNTIME__PORT", os.environ.get("BLOOM_PORT", str(default_port))
        )
    )
    return host, port


def _deployment_shared_certs_dir() -> Path:
    from bloom_lims.config import _resolve_deployment_code

    return shared_dayhoff_certs_dir(_resolve_deployment_code())


def _write_runtime_meta(*, ssl_enabled: bool) -> None:
    _runtime_meta_file().write_text(
        json.dumps({"ssl_enabled": ssl_enabled}, sort_keys=True),
        encoding="utf-8",
    )


def _read_runtime_meta() -> dict[str, object]:
    meta_file = _runtime_meta_file()
    if not meta_file.exists():
        return {}
    try:
        payload = json.loads(meta_file.read_text(encoding="utf-8"))
    except Exception:
        return {}
    return payload if isinstance(payload, dict) else {}


def _clear_runtime_meta() -> None:
    _runtime_meta_file().unlink(missing_ok=True)


def _runtime_scheme() -> str:
    meta = _read_runtime_meta()
    if str(meta.get("ssl_enabled")).lower() in {"false", "0", "no"}:
        return "http"
    if str(meta.get("ssl_enabled")).lower() in {"true", "1", "yes"}:
        return "https"
    return "https"


@server_app.command("start")
def start(
    port: int = typer.Option(
        DEFAULT_BLOOM_WEB_PORT, "--port", "-p", help="Port to run on"
    ),
    host: str = typer.Option("0.0.0.0", "--host", "-h", help="Host to bind to"),  # nosec B104
    reload: bool = typer.Option(
        False, "--reload", "-r", help="Enable auto-reload for development"
    ),
    background: bool = typer.Option(
        True,
        "--background/--foreground",
        "-b/-f",
        help="Run in background",
    ),
    ssl: bool = typer.Option(
        True,
        "--ssl/--no-ssl",
        help="Serve over HTTPS with deployment-scoped certs",
    ),
    cert: str | None = typer.Option(None, "--cert", help="TLS certificate file"),
    key: str | None = typer.Option(None, "--key", help="TLS private key file"),
) -> None:
    """Start the BLOOM web UI."""
    _ensure_dir()
    host, port = _runtime_host_and_port(port, host)
    shown_host = display_host(host)
    if not ssl and (cert or key):
        console.print("[red]✗[/red] --cert and --key require HTTPS; omit them with --no-ssl")
        raise typer.Exit(1)

    protocol = "https" if ssl else "http"
    cert_file = key_file = None
    if ssl:
        resolved = resolve_https_certs(
            cert_path=cert,
            key_path=key,
            shared_certs_dir=_deployment_shared_certs_dir(),
            fallback_certs_dir=PROJECT_ROOT / "certs",
        )
        cert_file = resolved.cert_path
        key_file = resolved.key_path

    try:
        settings = get_settings()
        apply_runtime_environment(settings)
    except Exception as exc:
        console.print(f"[red]✗[/red] Failed to apply runtime environment: {exc}")
        raise typer.Exit(1) from exc

    preflight_errors: list[str] = []

    try:
        get_tapdb_db_config()
    except Exception as exc:
        preflight_errors.append(f"TapDB config: {exc}")

    auth = settings.auth
    if not auth.cognito_user_pool_id:
        preflight_errors.append("auth.cognito_user_pool_id is empty")
    if not auth.cognito_client_id:
        preflight_errors.append("auth.cognito_client_id is empty")
    if not auth.cognito_domain:
        preflight_errors.append("auth.cognito_domain is empty")
    if not auth.cognito_redirect_uri:
        preflight_errors.append("auth.cognito_redirect_uri is empty")

    if preflight_errors:
        console.print("[red]✗[/red] Startup aborted - missing required configuration:")
        for err in preflight_errors:
            console.print(f"   • {err}")
        console.print(
            "\n   Fix your config ([cyan]bloom config edit[/cyan]) or run [cyan]bloom db init[/cyan]."
        )
        raise typer.Exit(1)

    atlas_secret_warning = atlas_webhook_secret_warning(settings)
    if atlas_secret_warning:
        console.print(f"[yellow]⚠[/yellow] {atlas_secret_warning}")
        console.print(
            "   Configure it in [cyan]~/.config/bloom/config.yaml[/cyan] under "
            "[cyan]atlas.webhook_secret[/cyan]."
        )
        console.log(atlas_secret_warning)

    pid, _ = active_server_pid()
    if pid:
        console.print(f"[yellow]⚠[/yellow] Server already running (PID {pid})")
        console.print(f"   URL: [cyan]{protocol}://{shown_host}:{port}[/cyan]")
        console.print(
            "   Use [cyan]bloom server stop[/cyan] to stop or [cyan]bloom server logs[/cyan] to view logs"
        )
        return

    cmd = [
        sys.executable,
        "-m",
        "uvicorn",
        "main:app",
        "--host",
        host,
        "--port",
        str(port),
    ]
    if reload:
        cmd.append("--reload")
    if ssl and cert_file and key_file:
        cmd.extend(["--ssl-keyfile", str(key_file), "--ssl-certfile", str(cert_file)])

    env = os.environ.copy()
    env["PYTHONUNBUFFERED"] = "1"
    _write_runtime_meta(ssl_enabled=ssl)

    if background:
        log_file = new_log_path(_log_dir())
        with open(log_file, "w", buffering=1) as log_f:
            proc = subprocess.Popen(
                cmd,
                stdout=log_f,
                stderr=subprocess.STDOUT,
                start_new_session=True,
                cwd=PROJECT_ROOT,
                env=env,
            )
            time.sleep(2)
            if proc.poll() is not None:
                _clear_runtime_meta()
                console.print("[red]✗[/red] Server failed to start. Check logs:")
                console.print(f"   [dim]{log_file}[/dim]")
                raise typer.Exit(1)
            write_pid(_pid_file(), proc.pid)
            console.print(f"[green]✓[/green] Server started (PID {proc.pid})")
            console.print(f"   URL: [cyan]{protocol}://{shown_host}:{port}[/cyan]")
            console.print(f"   Logs: [dim]{log_file}[/dim]")
            console.print(
                "   Use [cyan]bloom server logs[/cyan] to view logs or [cyan]bloom server stop[/cyan] to stop"
            )
        return

    console.print(f"[cyan]Starting BLOOM UI on {host}:{port}...[/cyan]")
    console.print(f"   URL: [cyan]{protocol}://{shown_host}:{port}[/cyan]")
    console.print("   Press Ctrl+C to stop\n")
    try:
        subprocess.run(cmd, env=env, cwd=PROJECT_ROOT)
    except KeyboardInterrupt:
        console.print("\n[yellow]⚠[/yellow] Server stopped")
    except Exception:
        _clear_runtime_meta()
        raise


@server_app.command("stop")
def stop() -> None:
    """Stop the BLOOM web UI."""
    stopped, msg = stop_pid(_pid_file())
    if stopped:
        _clear_runtime_meta()
        console.print(f"[green]✓[/green] {msg}")
    elif "Permission" in msg:
        console.print(f"[red]✗[/red] {msg}")
        raise typer.Exit(1)
    else:
        console.print(f"[yellow]⚠[/yellow] {msg}")


@server_app.command("status")
def status() -> None:
    """Show BLOOM server runtime status."""
    pid, _ = active_server_pid()
    host, port = _runtime_host_and_port(DEFAULT_BLOOM_WEB_PORT, "0.0.0.0")  # nosec B104
    shown_host = display_host(host)
    log_file = _latest_server_log()
    if pid:
        console.print(f"[green]●[/green] Server is [green]running[/green] (PID {pid})")
        console.print(f"   URL: [cyan]{_runtime_scheme()}://{shown_host}:{port}[/cyan]")
        if log_file:
            console.print(f"   Logs: [dim]{log_file}[/dim]")
        return

    console.print("[dim]○[/dim] Server is [dim]not running[/dim]")


@server_app.command("logs")
def logs(
    lines: int = typer.Option(50, "--lines", "-n", help="Number of lines to show"),
    follow: bool = typer.Option(False, "--follow", "-f", help="Follow log output"),
    service: LogService = typer.Option(
        LogService.all, "--service", "-s", help="Service to show logs for"
    ),
) -> None:
    """View BLOOM and TapDB operation logs."""
    log_files: list[tuple[str, Path]] = []

    if service in {LogService.server, LogService.all}:
        server_log = _latest_server_log()
        if server_log:
            log_files.append(("Server", server_log))

    if service in {LogService.tapdb, LogService.all}:
        tapdb_log = TAPDB_LOG_DIR / "db_operations.log"
        if tapdb_log.exists():
            log_files.append(("TapDB", tapdb_log))

    if not log_files:
        console.print("[yellow]No log files found.[/yellow]")
        console.print(f"  • Server: [cyan]{_log_dir()}/server_*.log[/cyan]")
        console.print("  • TapDB: [cyan]~/.config/tapdb/logs/db_operations.log[/cyan]")
        return

    for name, log_file in log_files:
        console.print(f"[bold]{name} Logs[/bold]: {log_file}")
        console.print()
        if follow:
            console.print("[dim](Press Ctrl+C to stop)[/dim]")
            try:
                subprocess.run(["tail", "-f", "-n", str(lines), str(log_file)])
            except KeyboardInterrupt:
                console.print()
        else:
            subprocess.run(["tail", "-n", str(lines), str(log_file)])


def register(registry: CommandRegistry, spec: CliSpec) -> None:
    """cli-core-yo plugin: register the server command group."""
    registry.add_typer_app(None, server_app, "server", "Server lifecycle commands.")



"""Bloom-specific user commands.

Generic user CRUD is owned by the TapDB CLI.
"""

from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from cli_core_yo.registry import CommandRegistry
    from cli_core_yo.spec import CliSpec

import typer
from daylily_tapdb.user_store import get_by_login_or_email

from bloom_lims.auth.rbac import API_ACCESS_GROUP, Role
from bloom_lims.auth.services.user_api_tokens import (
    TokenCreateInput,
    UserAPITokenService,
)
from bloom_lims.db import BLOOMdb3

users_app = typer.Typer(
    help="Bloom-specific user commands.", no_args_is_help=True
)


class TokenScope(str, Enum):
    internal_ro = "internal_ro"
    internal_rw = "internal_rw"
    admin = "admin"


@users_app.command("issue-token")
def issue_token(
    username: str = typer.Option(..., "--username", help="Username / login email"),
    token_name: str = typer.Option(
        ..., "--token-name", help="Display name for the token"
    ),
    scope: TokenScope = typer.Option(TokenScope.admin, "--scope", help="Token scope"),
    expires_in_days: int = typer.Option(
        30,
        "--expires-in-days",
        help="Token validity period in days",
    ),
    note: str = typer.Option("", "--note", help="Optional token note"),
) -> None:
    """Issue a user API token in the active Bloom namespace."""
    bdb = BLOOMdb3(app_username="cli-users")
    try:
        owner = get_by_login_or_email(bdb.session, username, include_inactive=True)
        if owner is None:
            typer.echo(f"User not found: {username}", err=True)
            raise typer.Exit(1)

        service = UserAPITokenService(bdb.session)
        service.groups.ensure_system_groups()
        created = service.create_token(
            owner_user_id=str(owner.uid),
            actor_user_id=str(owner.uid),
            actor_roles=[Role.ADMIN.value],
            actor_groups=[Role.ADMIN.value, API_ACCESS_GROUP],
            payload=TokenCreateInput(
                token_name=token_name.strip(),
                scope=scope.value,
                expires_in_days=expires_in_days,
                note=note.strip() or None,
            ),
        )
        bdb.session.commit()
    except PermissionError as exc:
        bdb.session.rollback()
        typer.echo(str(exc), err=True)
        raise typer.Exit(1) from exc
    finally:
        bdb.close()

    typer.echo(f"token_id={created.token.id}")
    typer.echo(f"token_prefix={created.token.token_prefix}")
    typer.echo(f"scope={created.token.scope}")
    typer.echo(f"plaintext_token={created.plaintext_token}")


def register(registry: CommandRegistry, spec: CliSpec) -> None:
    """cli-core-yo plugin: register the users command group."""
    registry.add_typer_app(None, users_app, "users", "User management commands.")



"""Typed zebra_day integration wrapper for Bloom."""

from __future__ import annotations

import importlib
import logging
import sys
from pathlib import Path
from typing import Any

import httpx

try:
    from zebra_day.client import PrinterRecord, ZebraDayApiClient
except ImportError:  # pragma: no cover - local sibling checkout fallback
    sibling_repo = Path(__file__).resolve().parents[4] / "zebra_day"
    if sibling_repo.exists():
        sys.path.insert(0, str(sibling_repo))
        sys.modules.pop("zebra_day", None)
    client_mod = importlib.import_module("zebra_day.client")
    PrinterRecord = client_mod.PrinterRecord
    ZebraDayApiClient = client_mod.ZebraDayApiClient

from bloom_lims.config import BloomSettings, get_settings


logger = logging.getLogger(__name__)


class ZebraDayIntegrationError(RuntimeError):
    """Raised when Bloom cannot safely use the configured zebra_day service."""


def _display_name(printer: PrinterRecord) -> str:
    return str(printer.printer_name or "").strip() or printer.printer_id


class ZebraDayService:
    """Thin service wrapper around the zebra_day API-backed client."""

    def __init__(self, settings: BloomSettings | None = None):
        self.settings = settings or get_settings()
        self.config = self.settings.zebra_day

    @property
    def base_url(self) -> str:
        return str(self.config.base_url or "").strip().rstrip("/")

    @property
    def admin_url(self) -> str | None:
        return self.base_url or None

    @property
    def is_configured(self) -> bool:
        return bool(self.base_url)

    @property
    def has_service_token(self) -> bool:
        return bool(str(self.config.token or "").strip())

    def _client(self) -> ZebraDayApiClient:
        if not self.is_configured:
            raise ZebraDayIntegrationError("zebra_day.base_url is not configured")

        headers = {"accept": "application/json"}
        if self.has_service_token:
            headers["authorization"] = f"Bearer {self.config.token}"

        http_client = httpx.Client(
            base_url=self.base_url,
            headers=headers,
            follow_redirects=True,
            timeout=float(self.config.timeout_seconds),
            verify=self.config.verify_ssl,
        )
        return ZebraDayApiClient(
            base_url=self.base_url,
            api_key=self.config.token or None,
            verify_ssl=self.config.verify_ssl,
            client=http_client,
        )

    def list_labs(self) -> list[str]:
        with self._client() as client:
            return list(client.list_labs())

    def list_printers(self, lab: str) -> list[PrinterRecord]:
        with self._client() as client:
            return list(client.list_printers(lab))

    def list_label_profiles(self) -> list[dict[str, Any]]:
        with self._client() as client:
            return list(client.list_label_profiles())

    def build_printer_preferences(self, selected_lab: str | None = None) -> dict[str, Any]:
        if not self.is_configured:
            return {
                "status": "not_configured",
                "error_detail": "zebra_day.base_url is not configured",
                "print_lab": [],
                "printer_name": [],
                "printer_options": [],
                "label_zpl_style": [],
                "selected_lab": "",
            }

        try:
            labs = self.list_labs()
            active_lab = str(selected_lab or "").strip()
            if active_lab not in labs:
                active_lab = labs[0] if labs else ""
            printers = self.list_printers(active_lab) if active_lab else []
            profiles = self.list_label_profiles()
        except Exception as exc:
            logger.warning("Failed to load zebra_day printer preferences: %s", exc)
            return {
                "status": "error",
                "error_detail": str(exc),
                "print_lab": [],
                "printer_name": [],
                "printer_options": [],
                "label_zpl_style": [],
                "selected_lab": "",
            }

        printer_options = [
            {
                "value": printer.printer_id,
                "label": _display_name(printer),
                "lab": printer.lab,
                "ip_address": printer.ip_address,
            }
            for printer in printers
        ]
        label_profiles = sorted(
            {
                str(item.get("profile_name") or "").strip()
                for item in profiles
                if str(item.get("profile_name") or "").strip()
            }
        )

        return {
            "status": "ok",
            "error_detail": "",
            "print_lab": labs,
            "printer_name": [item["value"] for item in printer_options],
            "printer_options": printer_options,
            "label_zpl_style": label_profiles,
            "selected_lab": active_lab,
        }

    def submit_print_job(
        self,
        *,
        lab: str,
        printer_id: str,
        label_zpl_style: str,
        euid: str,
        alt_a: str = "",
        alt_b: str = "",
        alt_c: str = "",
        alt_d: str = "",
        alt_e: str = "",
        alt_f: str = "",
        print_n: int = 1,
    ) -> dict[str, Any]:
        with self._client() as client:
            return dict(
                client.submit_print_job(
                    lab=lab,
                    printer=printer_id,
                    label_zpl_style=label_zpl_style or None,
                    uid_barcode=euid,
                    alt_a=alt_a,
                    alt_b=alt_b,
                    alt_c=alt_c,
                    alt_d=alt_d,
                    alt_e=alt_e,
                    alt_f=alt_f,
                    copies=int(print_n or 1),
                )
            )



from __future__ import annotations

import asyncio

from bloom_lims.api.rate_limiting import RateLimiter, RateLimitMiddleware


async def _ok_app(scope, receive, send):
    await send({"type": "http.response.start", "status": 200, "headers": []})
    await send({"type": "http.response.body", "body": b"ok"})


async def _run_request(path: str, middleware: RateLimitMiddleware):
    messages: list[dict] = []

    async def receive():
        return {"type": "http.request", "body": b"", "more_body": False}

    async def send(message):
        messages.append(message)

    scope = {
        "type": "http",
        "path": path,
        "method": "GET",
        "headers": [],
        "query_string": b"",
        "client": ("127.0.0.1", 12345),
    }
    await middleware(scope, receive, send)
    return messages


def _status(messages: list[dict]) -> int:
    return next(
        msg["status"] for msg in messages if msg["type"] == "http.response.start"
    )


def test_rate_limit_middleware_limits_untrusted_paths():
    limiter = RateLimiter(requests_per_minute=1, requests_per_hour=1000, burst_size=1)
    limiter.requests_per_minute = 1
    limiter.requests_per_hour = 1000
    limiter.burst_size = 1
    limiter.reset()

    allowed_1, _info_1 = limiter.is_allowed("127.0.0.1")
    allowed_2, info_2 = limiter.is_allowed("127.0.0.1")

    assert allowed_1 is True
    assert allowed_2 is False
    assert info_2["rate_limited"] is True


def test_rate_limit_middleware_skips_atlas_external_paths():
    middleware = RateLimitMiddleware(
        _ok_app,
        requests_per_minute=1,
        requests_per_hour=1000,
    )

    first = asyncio.run(_run_request("/api/v1/external/atlas/beta/tubes", middleware))
    second = asyncio.run(
        _run_request("/api/v1/external/atlas/beta/materials", middleware)
    )

    assert _status(first) == 200
    assert _status(second) == 200



from __future__ import annotations

from dataclasses import dataclass
from datetime import UTC, datetime
from typing import Any

from daylily_tapdb import require_seeded_templates
from daylily_tapdb.factory import InstanceFactory
from daylily_tapdb.models.instance import generic_instance
from daylily_tapdb.templates import TemplateManager
from sqlalchemy import select
from sqlalchemy.orm import Session
from sqlalchemy.orm.attributes import flag_modified

from bloom_lims.config import get_settings
from bloom_lims.observability import ProjectionMetadata

ANOMALY_TEMPLATE_CODE = "bloom/ops/anomaly-record/1.0/"
ANOMALY_PREFIX = "BAN"


@dataclass(frozen=True)
class AnomalyRecord:
    id: str
    service: str
    environment: str
    category: str
    severity: str
    fingerprint: str
    summary: str
    first_seen_at: str
    last_seen_at: str
    occurrence_count: int
    redacted_context: dict[str, Any]
    source_view_url: str


def redact_context(value: Any) -> Any:
    if isinstance(value, dict):
        redacted: dict[str, Any] = {}
        for key, item in value.items():
            lowered = str(key).lower()
            if lowered in {"authorization", "cookie", "set-cookie", "token", "secret", "password"}:
                redacted[str(key)] = "[redacted]"
            else:
                redacted[str(key)] = redact_context(item)
        return redacted
    if isinstance(value, list):
        return [redact_context(item) for item in value]
    if isinstance(value, str):
        text = value
        for prefix in ("/documents/", "/patients/", "/subjects/"):
            if prefix in text:
                head, _, tail = text.partition(prefix)
                parts = tail.split("/", 1)
                if parts and parts[0].isdigit():
                    suffix = f"/{parts[1]}" if len(parts) > 1 else ""
                    text = f"{head}{prefix}{{id}}{suffix}"
        if "select " in text.lower():
            return "[redacted-sql]"
        return text
    return value


class TapdbAnomalyRepository:
    def __init__(self, db: Session):
        self.db = db
        self.templates = TemplateManager()
        self.factory = InstanceFactory(self.templates)
        self._templates_ready = False

    def projection(self, *, observed_at: str | None = None) -> ProjectionMetadata:
        seen_at = observed_at or datetime.now(UTC).isoformat()
        return ProjectionMetadata(
            state="ready",
            stale=False,
            observed_at=seen_at,
            last_synced_at=seen_at,
            detail=None,
        )

    def record(
        self,
        *,
        category: str,
        severity: str,
        fingerprint: str,
        summary: str,
        redacted_context: dict[str, Any] | None = None,
    ) -> AnomalyRecord:
        self._ensure_templates()
        now = datetime.now(UTC).isoformat()
        environment = get_settings().environment or "unknown"
        context = redact_context(redacted_context or {})
        existing = self._find_existing(
            category=category,
            severity=severity,
            fingerprint=fingerprint,
            environment=environment,
        )
        if existing is None:
            instance = self.factory.create_instance(
                session=self.db,
                template_code=ANOMALY_TEMPLATE_CODE,
                name=summary[:120],
                properties={
                    "service": "bloom",
                    "environment": environment,
                    "category": category,
                    "severity": severity,
                    "fingerprint": fingerprint,
                    "summary": summary,
                    "first_seen_at": now,
                    "last_seen_at": now,
                    "occurrence_count": 1,
                    "redacted_context": context,
                },
            )
        else:
            props = self._props(existing)
            props["summary"] = summary
            props["last_seen_at"] = now
            props["occurrence_count"] = int(props.get("occurrence_count") or 0) + 1
            props["redacted_context"] = context
            self._write_props(existing, props)
            instance = existing
        self.db.commit()
        return self._to_record(instance)

    def list(self, *, skip: int = 0, limit: int = 100) -> list[AnomalyRecord]:
        self._ensure_templates()
        rows = [self._to_record(instance) for instance in self._instances()]
        rows.sort(key=lambda row: row.last_seen_at, reverse=True)
        return rows[skip : skip + limit]

    def get(self, anomaly_id: str) -> AnomalyRecord | None:
        self._ensure_templates()
        for instance in self._instances():
            if instance.euid == anomaly_id:
                return self._to_record(instance)
        return None

    def record_db_probe_failure(self, *, detail: str, latency_ms: float) -> AnomalyRecord:
        return self.record(
            category="database",
            severity="error",
            fingerprint=f"db:{hash(detail)}",
            summary="Bloom database probe failed",
            redacted_context={
                "detail": detail,
                "latency_ms": round(float(latency_ms), 3),
            },
        )

    def _ensure_templates(self) -> None:
        if self._templates_ready:
            return
        require_seeded_templates(
            self.db,
            [(ANOMALY_TEMPLATE_CODE, ANOMALY_PREFIX)],
            app_name="Bloom",
            template_manager=self.templates,
        )
        self._templates_ready = True

    def _find_existing(
        self,
        *,
        category: str,
        severity: str,
        fingerprint: str,
        environment: str,
    ) -> generic_instance | None:
        for instance in self._instances():
            props = self._props(instance)
            if (
                str(props.get("category") or "") == category
                and str(props.get("severity") or "") == severity
                and str(props.get("fingerprint") or "") == fingerprint
                and str(props.get("environment") or "") == environment
            ):
                return instance
        return None

    def _instances(self) -> list[generic_instance]:
        category, type_name, subtype, version = ANOMALY_TEMPLATE_CODE.strip("/").split("/")
        stmt = (
            select(generic_instance)
            .where(
                generic_instance.category == category,
                generic_instance.type == type_name,
                generic_instance.subtype == subtype,
                generic_instance.version == version,
                generic_instance.is_deleted.is_(False),
            )
            .order_by(generic_instance.created_dt.desc())
        )
        return list(self.db.execute(stmt).scalars())

    def _props(self, instance: generic_instance) -> dict[str, Any]:
        payload = instance.json_addl or {}
        if not isinstance(payload, dict):
            payload = {}
        properties = payload.get("properties")
        if not isinstance(properties, dict):
            properties = {}
            payload["properties"] = properties
            instance.json_addl = payload
        return properties

    def _write_props(self, instance: generic_instance, properties: dict[str, Any]) -> None:
        payload = instance.json_addl or {}
        if not isinstance(payload, dict):
            payload = {}
        payload["properties"] = properties
        instance.json_addl = payload
        if hasattr(instance, "_sa_instance_state"):
            flag_modified(instance, "json_addl")

    def _to_record(self, instance: generic_instance) -> AnomalyRecord:
        props = self._props(instance)
        return AnomalyRecord(
            id=str(instance.euid),
            service=str(props.get("service") or "bloom"),
            environment=str(props.get("environment") or (get_settings().environment or "unknown")),
            category=str(props.get("category") or "unknown"),
            severity=str(props.get("severity") or "unknown"),
            fingerprint=str(props.get("fingerprint") or ""),
            summary=str(props.get("summary") or ""),
            first_seen_at=str(props.get("first_seen_at") or ""),
            last_seen_at=str(props.get("last_seen_at") or ""),
            occurrence_count=int(props.get("occurrence_count") or 0),
            redacted_context=dict(props.get("redacted_context") or {}),
            source_view_url=f"/admin/anomalies/{instance.euid}",
        )


def open_anomaly_repository(*, app_username: str):
    from bloom_lims.db import BLOOMdb3

    bdb = BLOOMdb3(app_username=app_username)
    return TapdbAnomalyRepository(bdb.session), bdb



from __future__ import annotations

import importlib.metadata
import json
import os
import subprocess
import sys
from datetime import UTC, datetime
from pathlib import Path
from typing import Any

from bloom_lims.config import apply_runtime_environment, get_settings


def _utcnow() -> str:
    return datetime.now(UTC).isoformat()


def _state_dir() -> Path:
    try:
        from cli_core_yo.runtime import get_context

        return get_context().xdg_paths.state
    except Exception:
        return Path.home() / ".local" / "state" / "bloom"


def schema_drift_state_file() -> Path:
    return _state_dir() / "schema_drift.json"


def tapdb_tool_version() -> str:
    try:
        return importlib.metadata.version("daylily-tapdb")
    except importlib.metadata.PackageNotFoundError:
        return ""


def _tapdb_cmd(args: list[str]) -> subprocess.CompletedProcess[str]:
    env = apply_runtime_environment(get_settings())
    runtime = os.environ.copy()
    runtime["AWS_PROFILE"] = env.aws_profile
    runtime["AWS_REGION"] = env.aws_region
    runtime["AWS_DEFAULT_REGION"] = env.aws_region
    cmd = [
        sys.executable,
        "-m",
        "daylily_tapdb.cli",
        "--env",
        env.env,
    ]
    if not str(env.config_path or "").strip():
        raise RuntimeError("TapDB schema drift checks require an explicit config path.")
    cmd.extend(["--config", env.config_path])
    cmd.extend(args)
    return subprocess.run(
        cmd,
        env=runtime,
        capture_output=True,
        text=True,
    )


def _schema_drift_summary(payload: dict[str, object]) -> str:
    counts = payload.get("counts")
    if isinstance(counts, dict):
        expected = counts.get("expected")
        live = counts.get("live")
        return f"expected={expected} live={live}"
    return "drift report available"


def _status_for_returncode(returncode: int) -> str:
    if returncode == 0:
        return "clean"
    if returncode == 1:
        return "drift"
    return "check_failed"


def _truncate_stderr(stderr: str) -> str:
    cleaned = (stderr or "").strip()
    if len(cleaned) <= 1000:
        return cleaned
    return f"{cleaned[:1000]}..."


def run_schema_drift_check(env_name: str) -> dict[str, Any]:
    result = _tapdb_cmd(["db", "schema", "drift-check", env_name, "--json", "--no-strict"])
    payload: dict[str, object] = {}
    if result.stdout.strip():
        try:
            parsed = json.loads(result.stdout)
            if isinstance(parsed, dict):
                payload = parsed
        except json.JSONDecodeError:
            payload = {"raw_stdout": result.stdout.strip()}

    status = _status_for_returncode(result.returncode)
    return {
        "status": status,
        "checked_at": _utcnow(),
        "environment": env_name,
        "tool_version": tapdb_tool_version(),
        "summary": _schema_drift_summary(payload) if payload else ("schema drift check failed" if status == "check_failed" else "schema drift check completed"),
        "report": payload,
        "stderr": _truncate_stderr(result.stderr) if status == "check_failed" else "",
    }


def write_schema_drift_report(report: dict[str, Any]) -> None:
    path = schema_drift_state_file()
    path.parent.mkdir(parents=True, exist_ok=True)
    path.write_text(json.dumps(report, indent=2, sort_keys=True), encoding="utf-8")


def read_schema_drift_report(*, environment: str) -> dict[str, Any]:
    path = schema_drift_state_file()
    if not path.exists():
        return {
            "status": "not_run",
            "checked_at": None,
            "environment": environment,
            "tool_version": tapdb_tool_version(),
            "summary": "Schema drift check has not been run",
            "report": {},
            "stderr": "",
        }
    try:
        payload = json.loads(path.read_text(encoding="utf-8"))
    except Exception as exc:
        return {
            "status": "check_failed",
            "checked_at": _utcnow(),
            "environment": environment,
            "tool_version": tapdb_tool_version(),
            "summary": "Stored schema drift report could not be read",
            "report": {},
            "stderr": _truncate_stderr(str(exc)),
        }
    if not isinstance(payload, dict):
        return {
            "status": "check_failed",
            "checked_at": _utcnow(),
            "environment": environment,
            "tool_version": tapdb_tool_version(),
            "summary": "Stored schema drift report has invalid format",
            "report": {},
            "stderr": "",
        }
    payload.setdefault("status", "check_failed")
    payload.setdefault("checked_at", None)
    payload.setdefault("environment", environment)
    payload.setdefault("tool_version", tapdb_tool_version())
    payload.setdefault("summary", "")
    payload.setdefault("report", {})
    payload.setdefault("stderr", "")
    return payload



"""
Tests for BLOOM LIMS Pydantic Schemas

These tests verify that the Pydantic schemas work correctly for
validation, serialization, and deserialization.
"""

from datetime import datetime

import pytest


@pytest.fixture(autouse=True)
def _default_meridian_runtime(monkeypatch):
    monkeypatch.setenv("MERIDIAN_ENVIRONMENT", "production")
    monkeypatch.setenv("MERIDIAN_SANDBOX_PREFIX", "")


class TestBaseSchemas:
    """Tests for base schema functionality."""

    def test_pagination_params(self):
        """Test PaginationParams schema."""
        from bloom_lims.schemas import PaginationParams

        params = PaginationParams(page=1, page_size=50)
        assert params.page == 1
        assert params.page_size == 50

        # Test defaults
        params_default = PaginationParams()
        assert params_default.page == 1
        assert params_default.page_size == 50

    def test_euid_validation(self, monkeypatch):
        """Test EUID validation function.

        BLOOM uses TapDB/Meridian EUIDs:
        - Production: PREFIX-BODYCHECK (e.g. BCN-19, GT-13)
        - BODY is Crockford Base32 (no leading zeros)
        - CHECK is MOD32 checksum
        """
        from daylily_tapdb.euid import format_euid

        from bloom_lims.schemas import validate_euid

        cx1 = format_euid("BCN", 1)
        cx123 = format_euid("BCN", 123)
        wx1000 = format_euid("BWF", 1000)

        # Valid EUIDs (TapDB/Meridian)
        assert validate_euid(cx1) == cx1
        assert validate_euid(cx123) == cx123
        assert (
            validate_euid(f"  {wx1000.lower()}  ") == wx1000
        )  # Should uppercase and strip

        # Invalid EUIDs
        with pytest.raises(ValueError):
            validate_euid("")
        with pytest.raises(ValueError):
            validate_euid("   ")
        with pytest.raises(ValueError):
            validate_euid("CX1")  # Missing dash+checksum segment
        with pytest.raises(ValueError):
            validate_euid("BCN-01")  # Leading zeros not allowed in BODY

        # Bad checksum
        bad_checksum = cx1[:-1] + ("0" if cx1[-1] != "0" else "1")
        with pytest.raises(ValueError):
            validate_euid(bad_checksum)

        # Sandbox EUIDs are not accepted by Bloom's public schema validators
        sandbox = format_euid("BCN", 1, sandbox="X")
        with pytest.raises(ValueError):
            validate_euid(sandbox)

    def test_euid_validation_accepts_only_configured_sandbox_prefix(self, monkeypatch):
        monkeypatch.setenv("MERIDIAN_ENVIRONMENT", "sandbox")
        monkeypatch.setenv("MERIDIAN_SANDBOX_PREFIX", "S")

        from daylily_tapdb.euid import format_euid

        from bloom_lims.schemas import validate_euid

        sandbox = format_euid("BCN", 1, sandbox="S")
        wrong_prefix = format_euid("BCN", 1, sandbox="T")
        production = format_euid("BCN", 1)

        assert validate_euid(sandbox) == sandbox
        with pytest.raises(ValueError):
            validate_euid(wrong_prefix)
        with pytest.raises(ValueError):
            validate_euid(production)


class TestObjectSchemas:
    """Tests for object schemas."""

    def test_object_create_schema(self):
        """Test ObjectCreateSchema validation."""
        from bloom_lims.schemas import ObjectCreateSchema

        data = ObjectCreateSchema(
            name="Test Object",
            type="sample",
            subtype="blood",
        )
        assert data.name == "Test Object"
        assert data.type == "sample"
        assert data.subtype == "blood"

    def test_object_create_with_json_addl(self):
        """Test ObjectCreateSchema with json_addl."""
        from bloom_lims.schemas import ObjectCreateSchema

        data = ObjectCreateSchema(
            name="Test Object",
            type="sample",
            json_addl={"properties": {"key": "value"}},
        )
        assert data.json_addl == {"properties": {"key": "value"}}


class TestContainerSchemas:
    """Tests for container schemas."""

    def test_container_create_schema(self):
        """Test ContainerCreateSchema validation."""
        from daylily_tapdb.euid import format_euid

        from bloom_lims.schemas import ContainerCreateSchema

        data = ContainerCreateSchema(
            name="Test Plate",
            container_type="plate",
            subtype="96-well",
            template_euid=format_euid("GT", 1),
        )
        assert data.name == "Test Plate"
        assert data.container_type == "plate"

    def test_container_layout_schema(self):
        """Test ContainerLayoutSchema."""
        from bloom_lims.schemas import ContainerLayoutSchema

        layout = ContainerLayoutSchema(
            rows=8,
            columns=12,
            layout_type="grid",
        )
        assert layout.rows == 8
        assert layout.columns == 12


class TestContentSchemas:
    """Tests for content schemas."""

    def test_sample_create_schema(self):
        """Test SampleCreateSchema validation."""
        from daylily_tapdb.euid import format_euid

        from bloom_lims.schemas import SampleCreateSchema

        data = SampleCreateSchema(
            name="Test Sample",
            sample_type="blood",
            template_euid=format_euid("GT", 2),
        )
        assert data.name == "Test Sample"
        assert data.sample_type == "blood"

    def test_reagent_create_schema(self):
        """Test ReagentCreateSchema validation."""
        from daylily_tapdb.euid import format_euid

        from bloom_lims.schemas import ReagentCreateSchema

        data = ReagentCreateSchema(
            name="Test Reagent",
            reagent_type="buffer",
            template_euid=format_euid("GT", 3),
            lot_number="LOT001",
        )
        assert data.lot_number == "LOT001"


class TestWorkflowSchemas:
    """Tests for workflow schemas."""

    def test_workflow_create_schema(self):
        """Test WorkflowCreateSchema validation."""
        from daylily_tapdb.euid import format_euid

        from bloom_lims.schemas import WorkflowCreateSchema

        data = WorkflowCreateSchema(
            name="Test Workflow",
            workflow_type="sequencing",
            template_euid=format_euid("GT", 4),
        )
        assert data.name == "Test Workflow"

    def test_workflow_step_schema(self):
        """Test WorkflowStepSchema validation."""
        from bloom_lims.schemas import WorkflowStepSchema

        step = WorkflowStepSchema(
            name="Step 1",
            step_number=1,
            step_type="extraction",
        )
        assert step.step_number == 1


class TestEquipmentSchemas:
    """Tests for equipment schemas."""

    def test_equipment_create_schema(self):
        """Test EquipmentCreateSchema validation."""
        from bloom_lims.schemas import EquipmentCreateSchema

        data = EquipmentCreateSchema(
            name="Test Sequencer",
            equipment_type="sequencer",
            serial_number="SN12345",
            manufacturer="Illumina",
        )
        assert data.serial_number == "SN12345"

    def test_maintenance_record_schema(self):
        """Test MaintenanceRecordSchema validation."""
        from bloom_lims.schemas import MaintenanceRecordSchema

        record = MaintenanceRecordSchema(
            maintenance_type="preventive",
            performed_date=datetime.now(),
            performed_by="Tech1",
        )
        assert record.maintenance_type == "preventive"


class TestPublicResponseSchemas:
    """Tests for public response schema contracts."""

    def test_public_response_schemas_do_not_expose_uuid_fields(self):
        from bloom_lims.schemas.actions import ActionResponseSchema
        from bloom_lims.schemas.containers import (
            ContainerPositionSchema,
            ContainerResponseSchema,
        )
        from bloom_lims.schemas.content import ContentResponseSchema
        from bloom_lims.schemas.equipment import EquipmentResponseSchema
        from bloom_lims.schemas.files import FileResponseSchema, FileSetResponseSchema
        from bloom_lims.schemas.objects import ObjectResponseSchema
        from bloom_lims.schemas.workflows import (
            WorkflowResponseSchema,
            WorkflowStepResponseSchema,
        )

        schemas = [
            ActionResponseSchema,
            ContainerPositionSchema,
            ContainerResponseSchema,
            ContentResponseSchema,
            EquipmentResponseSchema,
            FileResponseSchema,
            FileSetResponseSchema,
            ObjectResponseSchema,
            WorkflowResponseSchema,
            WorkflowStepResponseSchema,
        ]

        for schema in schemas:
            assert "uuid" not in schema.model_fields


class TestDomainUtils:
    """Tests for bloom_lims.domain.utils module."""

    def test_get_clean_timestamp(self):
        """Test clean timestamp generation."""
        from bloom_lims.domain.utils import get_clean_timestamp

        ts = get_clean_timestamp()
        assert isinstance(ts, str)
        # Format should be YYYY-MM-DD_HH-MM-SS
        assert len(ts) == 19
        assert ts[4] == "-"
        assert ts[7] == "-"
        assert ts[10] == "_"

    def test_generate_random_string_default_length(self):
        """Test random string generation with default length."""
        from bloom_lims.domain.utils import generate_random_string

        s = generate_random_string()
        assert isinstance(s, str)
        assert len(s) == 10
        # Should only contain alphanumeric characters
        assert s.isalnum()

    def test_generate_random_string_custom_length(self):
        """Test random string generation with custom length."""
        from bloom_lims.domain.utils import generate_random_string

        s = generate_random_string(20)
        assert len(s) == 20

        s2 = generate_random_string(5)
        assert len(s2) == 5

    def test_get_datetime_string(self):
        """Test datetime string generation."""
        from bloom_lims.domain.utils import get_datetime_string

        dt = get_datetime_string()
        assert isinstance(dt, str)
        # Should contain date, time, and timezone info
        assert "-" in dt  # Date separators
        assert ":" in dt  # Time separators

    def test_update_recursive_simple(self):
        """Test recursive dictionary update with simple values."""
        from bloom_lims.domain.utils import update_recursive

        orig = {"a": 1, "b": 2}
        update = {"b": 3, "c": 4}
        update_recursive(orig, update)
        assert orig == {"a": 1, "b": 3, "c": 4}

    def test_update_recursive_nested(self):
        """Test recursive dictionary update with nested dicts."""
        from bloom_lims.domain.utils import update_recursive

        orig = {"a": {"x": 1, "y": 2}, "b": 3}
        update = {"a": {"y": 20, "z": 30}}
        update_recursive(orig, update)
        assert orig == {"a": {"x": 1, "y": 20, "z": 30}, "b": 3}

    def test_unique_non_empty_strings(self):
        """Test unique non-empty string filtering."""
        from bloom_lims.domain.utils import unique_non_empty_strings

        # Should remove duplicates and empty strings
        result = unique_non_empty_strings(["a", "b", "", "a", "c", ""])
        assert set(result) == {"a", "b", "c"}

    def test_unique_non_empty_strings_all_empty(self):
        """Test unique non-empty strings with all empty."""
        from bloom_lims.domain.utils import unique_non_empty_strings

        result = unique_non_empty_strings(["", "", ""])
        assert result == []


class TestBloomExceptions:
    """Tests for bloom_lims.exceptions module.

    Note: Exception tests are skipped because exceptions log on creation
    and the logger 'extra' dict contains 'message' which conflicts with
    Python's built-in LogRecord 'message' attribute.
    """

    @pytest.mark.skip(
        reason="Logger extra contains 'message' key that conflicts with LogRecord"
    )
    def test_bloom_error_basic(self):
        """Test basic BloomError creation."""
        from bloom_lims.exceptions import BloomError

        error = BloomError("Test error message")
        assert error.message == "Test error message"

    @pytest.mark.skip(
        reason="Logger extra contains 'message' key that conflicts with LogRecord"
    )
    def test_validation_error(self):
        """Test ValidationError subclass."""
        from bloom_lims.exceptions import ValidationError

        error = ValidationError("Invalid input", field="email")
        assert error.http_status == 400


class TestExceptionImports:
    """Test that exception classes can be imported."""

    def test_import_bloom_error(self):
        """Test BloomError can be imported."""
        from bloom_lims.exceptions import BloomError

        assert BloomError is not None

    def test_import_validation_error(self):
        """Test ValidationError can be imported."""
        from bloom_lims.exceptions import ValidationError

        assert ValidationError is not None

    def test_import_not_found_error(self):
        """Test NotFoundError can be imported."""
        from bloom_lims.exceptions import NotFoundError

        assert NotFoundError is not None

    def test_import_database_error(self):
        """Test DatabaseError can be imported."""
        from bloom_lims.exceptions import DatabaseError

        assert DatabaseError is not None

    def test_import_authentication_error(self):
        """Test AuthenticationError can be imported."""
        from bloom_lims.exceptions import AuthenticationError

        assert AuthenticationError is not None

    def test_import_authorization_error(self):
        """Test AuthorizationError can be imported."""
        from bloom_lims.exceptions import AuthorizationError

        assert AuthorizationError is not None

    def test_import_configuration_error(self):
        """Test ConfigurationError can be imported."""
        from bloom_lims.exceptions import ConfigurationError

        assert ConfigurationError is not None

    def test_import_external_service_error(self):
        """Test ExternalServiceError can be imported."""
        from bloom_lims.exceptions import ExternalServiceError

        assert ExternalServiceError is not None


class TestDomainWorkflows:
    """Tests for workflow domain module imports and utilities."""

    def test_import_bloom_workflow(self):
        """Test BloomWorkflow import."""
        from bloom_lims.domain.workflows import BloomWorkflow

        assert BloomWorkflow is not None

    def test_import_bloom_workflow_step(self):
        """Test BloomWorkflowStep import."""
        from bloom_lims.domain.workflows import BloomWorkflowStep

        assert BloomWorkflowStep is not None

    def test_workflow_inherits_from_bloom_obj(self):
        """Test BloomWorkflow inherits from BloomObj."""
        from bloom_lims.domain.base import BloomObj
        from bloom_lims.domain.workflows import BloomWorkflow

        assert issubclass(BloomWorkflow, BloomObj)


class TestDomainFiles:
    """Tests for files domain module imports."""

    def test_import_bloom_file(self):
        """Test BloomFile import."""
        from bloom_lims.domain.files import BloomFile

        assert BloomFile is not None

    def test_import_bloom_file_set(self):
        """Test BloomFileSet import."""
        from bloom_lims.domain.files import BloomFileSet

        assert BloomFileSet is not None

    def test_file_inherits_from_bloom_obj(self):
        """Test BloomFile inherits from BloomObj."""
        from bloom_lims.domain.base import BloomObj
        from bloom_lims.domain.files import BloomFile

        assert issubclass(BloomFile, BloomObj)


class TestDomainContainers:
    """Tests for containers domain module imports."""

    def test_import_bloom_container(self):
        """Test BloomContainer import."""
        from bloom_lims.domain.containers import BloomContainer

        assert BloomContainer is not None


class TestDomainContent:
    """Tests for content domain module imports."""

    def test_import_bloom_content(self):
        """Test BloomContent import."""
        from bloom_lims.domain.content import BloomContent

        assert BloomContent is not None


class TestDomainEquipment:
    """Tests for equipment domain module imports."""

    def test_import_bloom_equipment(self):
        """Test BloomEquipment import."""
        from bloom_lims.domain.equipment import BloomEquipment

        assert BloomEquipment is not None


class TestCoreModuleImports:
    """Tests for core module imports."""

    def test_import_lru_cache(self):
        """Test LRUCache import."""
        from bloom_lims.core.cache import LRUCache

        assert LRUCache is not None

    def test_import_cache_entry(self):
        """Test CacheEntry import."""
        from bloom_lims.core.cache import CacheEntry

        assert CacheEntry is not None

    def test_import_cache_stats(self):
        """Test CacheStats import."""
        from bloom_lims.core.cache import CacheStats

        assert CacheStats is not None

    def test_import_template_validator(self):
        """Test TemplateValidator import."""
        from bloom_lims.core.template_validation import TemplateValidator

        assert TemplateValidator is not None

    def test_import_batch_processor(self):
        """Test BatchProcessor import."""
        from bloom_lims.core.batch_operations import BatchProcessor

        assert BatchProcessor is not None

    def test_import_async_operations(self):
        """Test async operations imports."""
        from bloom_lims.core.async_operations import BackgroundTaskManager, TaskStatus

        assert BackgroundTaskManager is not None
        assert TaskStatus is not None

    def test_import_task_result(self):
        """Test TaskResult import."""
        from bloom_lims.core.async_operations import TaskResult

        assert TaskResult is not None


class TestHealthModuleImports:
    """Tests for health module imports."""

    def test_import_health_router(self):
        """Test health_router import."""
        from bloom_lims.health import health_router

        assert health_router is not None

    def test_import_component_health(self):
        """Test ComponentHealth import."""
        from bloom_lims.health import ComponentHealth

        assert ComponentHealth is not None

    def test_import_health_response(self):
        """Test HealthResponse import."""
        from bloom_lims.health import HealthResponse

        assert HealthResponse is not None

    def test_import_liveness_response(self):
        """Test LivenessResponse import."""
        from bloom_lims.health import LivenessResponse

        assert LivenessResponse is not None

    def test_import_readiness_response(self):
        """Test ReadinessResponse import."""
        from bloom_lims.health import ReadinessResponse

        assert ReadinessResponse is not None

    def test_component_health_model(self):
        """Test ComponentHealth model creation."""
        from bloom_lims.health import ComponentHealth

        component = ComponentHealth(
            name="test", status="healthy", latency_ms=10.5, message="Test component"
        )
        assert component.name == "test"
        assert component.status == "healthy"
        assert component.latency_ms == 10.5

    def test_liveness_response_model(self):
        """Test LivenessResponse model creation."""
        from datetime import datetime

        from bloom_lims.health import LivenessResponse

        response = LivenessResponse(status="ok", timestamp=datetime.utcnow())
        assert response.status == "ok"
        assert response.timestamp is not None

    def test_get_system_info(self):
        """Test get_system_info function."""
        from bloom_lims.health import get_system_info

        info = get_system_info()
        assert isinstance(info, dict)
        # Should have system resource info
        if info:  # May be empty if psutil fails
            expected_keys = ["cpu_percent", "memory_percent", "python_version"]
            for key in expected_keys:
                if key in info:
                    assert info[key] is not None


class TestDomainBaseImports:
    """Tests for domain base module imports."""

    def test_import_bloom_obj(self):
        """Test BloomObj import."""
        from bloom_lims.domain.base import BloomObj

        assert BloomObj is not None

    def test_import_db_class(self):
        """Test BLOOMdb3 import."""
        from bloom_lims.db import BLOOMdb3

        assert BLOOMdb3 is not None


class TestConfigModuleTests:
    """Tests for config module."""

    def test_import_settings(self):
        """Test get_settings import."""
        from bloom_lims.config import get_settings

        assert get_settings is not None

    def test_get_settings_returns_settings(self):
        """Test get_settings returns a settings object."""
        from bloom_lims.config import get_settings

        settings = get_settings()
        assert settings is not None
        assert hasattr(settings, "environment")

    def test_settings_has_api_config(self):
        """Test settings has API config."""
        from bloom_lims.config import get_settings

        settings = get_settings()
        assert hasattr(settings, "api")
        assert hasattr(settings.api, "version")

    def test_settings_has_auth_config(self):
        """Test settings has auth config."""
        from bloom_lims.config import get_settings

        settings = get_settings()
        assert hasattr(settings, "auth")


class TestAPIVersioningImports:
    """Tests for API versioning imports."""

    def test_import_versioning(self):
        """Test versioning module import."""
        from bloom_lims.api import versioning

        assert versioning is not None


class TestRateLimitingImports:
    """Tests for rate limiting imports."""

    def test_import_rate_limiting(self):
        """Test rate limiting module import."""
        from bloom_lims.api import rate_limiting

        assert rate_limiting is not None

    def test_import_rate_limit_middleware(self):
        """Test RateLimitMiddleware import."""
        from bloom_lims.api.rate_limiting import RateLimitMiddleware

        assert RateLimitMiddleware is not None


class TestDomainBaseImportsExtended:
    """Extended tests for domain base module."""

    def test_import_bloom_obj_class(self):
        """Test BloomObj class import."""
        from bloom_lims.domain.base import BloomObj

        assert BloomObj is not None

    def test_import_bloomdb3_class(self):
        """Test BLOOMdb3 class import."""
        from bloom_lims.db import BLOOMdb3

        assert BLOOMdb3 is not None

    def test_bloom_obj_has_get_by_euid(self):
        """Test BloomObj has get_by_euid method."""
        from bloom_lims.domain.base import BloomObj

        assert hasattr(BloomObj, "get_by_euid")

    def test_bloom_obj_does_not_have_uuid_helpers(self):
        """Test BloomObj no longer exposes UUID helper methods."""
        from bloom_lims.domain.base import BloomObj

        assert not hasattr(BloomObj, "get")
        assert not hasattr(BloomObj, "delete_by_uuid")
        assert not hasattr(BloomObj, "create_instances_from_uuid")

    def test_bloom_obj_has_create_instance(self):
        """Test BloomObj has create_instance method."""
        from bloom_lims.domain.base import BloomObj

        assert hasattr(BloomObj, "create_instance")


class TestCoreExceptionsImports:
    """Tests for core exceptions module."""

    def test_import_bloom_error(self):
        """Test BloomError import."""
        from bloom_lims.core.exceptions import BloomError

        assert BloomError is not None

    def test_import_bloom_not_found_error(self):
        """Test BloomNotFoundError import."""
        from bloom_lims.core.exceptions import BloomNotFoundError

        assert BloomNotFoundError is not None

    def test_import_bloom_database_error(self):
        """Test BloomDatabaseError import."""
        from bloom_lims.core.exceptions import BloomDatabaseError

        assert BloomDatabaseError is not None

    def test_import_bloom_validation_error(self):
        """Test BloomValidationError import."""
        from bloom_lims.core.exceptions import BloomValidationError

        assert BloomValidationError is not None


class TestTapdbAdapterImports:
    """Tests for tapdb adapter module."""

    def test_import_tapdb_adapter(self):
        """Test tapdb adapter import."""
        from bloom_lims import tapdb_adapter

        assert tapdb_adapter is not None


class TestSubjectingImports:
    """Tests for subjecting module."""

    def test_import_subjecting(self):
        """Test subjecting module import."""
        from bloom_lims import subjecting

        assert subjecting is not None


class TestDomainObjectSetsImports:
    """Tests for domain object_sets module."""

    def test_import_object_sets(self):
        """Test object_sets module import."""
        from bloom_lims.domain import object_sets

        assert object_sets is not None


class TestBatchOperationsImports:
    """Tests for batch operations imports."""

    def test_import_batch_operations(self):
        """Test batch operations module import."""
        from bloom_lims.core import batch_operations

        assert batch_operations is not None

    def test_import_batch_processor(self):
        """Test BatchProcessor import."""
        from bloom_lims.core.batch_operations import BatchProcessor

        assert BatchProcessor is not None


class TestCacheBackendsImports:
    """Tests for cache backends imports."""

    def test_import_cache_backends(self):
        """Test cache backends module import."""
        from bloom_lims.core import cache_backends

        assert cache_backends is not None


class TestReadReplicasImports:
    """Tests for read replicas imports."""

    def test_import_read_replicas(self):
        """Test read replicas module import."""
        from bloom_lims.core import read_replicas

        assert read_replicas is not None


class TestLineageImports:
    """Tests for lineage imports."""

    def test_import_lineage(self):
        """Test lineage module import."""
        from bloom_lims.core import lineage

        assert lineage is not None


class TestCachedRepositoryImports:
    """Tests for cached repository imports."""

    def test_import_cached_repository(self):
        """Test cached repository module import."""
        from bloom_lims.core import cached_repository

        assert cached_repository is not None


class TestContainersImports:
    """Tests for core containers imports."""

    def test_import_core_containers(self):
        """Test core containers module import."""
        from bloom_lims.core import containers

        assert containers is not None


class TestContentImports:
    """Tests for core content imports."""

    def test_import_core_content(self):
        """Test core content module import."""
        from bloom_lims.core import content

        assert content is not None


class TestCoreWorkflowsImports:
    """Tests for core workflows imports."""

    def test_import_core_workflows(self):
        """Test core workflows module import."""
        from bloom_lims.core import workflows

        assert workflows is not None


class TestBaseObjectsImports:
    """Tests for base objects imports."""

    def test_import_base_objects(self):
        """Test base objects module import."""
        from bloom_lims.core import base_objects

        assert base_objects is not None



from dataclasses import dataclass
from functools import lru_cache
from typing import List, Optional
from urllib.parse import urlencode, urlparse

import httpx
import jwt


class CognitoConfigurationError(Exception):
    """Raised when required Cognito configuration is missing."""


class CognitoTokenError(Exception):
    """Raised when Cognito tokens cannot be validated."""


@dataclass(frozen=True)
class CognitoConfig:
    user_pool_id: str
    client_id: str
    client_secret: str
    region: str
    domain: str
    redirect_uri: str
    logout_redirect_uri: str
    scopes: List[str]

    @property
    def _bare_domain(self) -> str:
        """Domain without protocol prefix for URL construction."""
        d = (self.domain or "").strip().rstrip("/")
        if d.startswith("https://"):
            d = d[len("https://"):]
        elif d.startswith("http://"):
            d = d[len("http://"):]
        return d

    @property
    def issuer(self) -> str:
        return f"https://cognito-idp.{self.region}.amazonaws.com/{self.user_pool_id}"

    @property
    def jwks_url(self) -> str:
        return f"{self.issuer}/.well-known/jwks.json"

    @property
    def authorize_url(self) -> str:
        query = urlencode(
            {
                "client_id": self.client_id,
                "response_type": "code",
                "redirect_uri": self.redirect_uri,
                "scope": " ".join(self.scopes),
            }
        )
        return f"https://{self._bare_domain}/oauth2/authorize?{query}"

    @property
    def token_url(self) -> str:
        return f"https://{self._bare_domain}/oauth2/token"

    @property
    def logout_url(self) -> str:
        query = urlencode(
            {
                "client_id": self.client_id,
                "logout_uri": self.logout_redirect_uri.rstrip("/"),
            }
        )
        return f"https://{self._bare_domain}/logout?{query}"


class CognitoAuth:
    def __init__(self, config: CognitoConfig):
        self.config = config
        self._jwks_client = jwt.PyJWKClient(config.jwks_url)

    @classmethod
    def _extract_port(cls, url: str) -> Optional[int]:
        normalized = (url or "").strip()
        if not normalized:
            return None
        parsed = urlparse(normalized)
        if not parsed.scheme or not parsed.netloc:
            return None
        if parsed.port is not None:
            return parsed.port
        if parsed.scheme.lower() == "https":
            return 443
        if parsed.scheme.lower() == "http":
            return 80
        return None

    @classmethod
    def _validate_runtime_uri_port_match(
        cls,
        *,
        label: str,
        stored_url: str,
        expected_url: str,
        pool_id: str,
        client_id: str,
    ) -> None:
        """Validate that stored Cognito app URI port matches current Bloom runtime port."""
        if not expected_url:
            return

        if not stored_url:
            raise CognitoConfigurationError(
                f"Unable to verify Cognito {label} URL for pool '{pool_id}' client '{client_id}'. "
                f"No stored {label} URL found."
            )

        stored_port = cls._extract_port(stored_url)
        expected_port = cls._extract_port(expected_url)
        if stored_port is None:
            raise CognitoConfigurationError(
                f"Stored Cognito {label} URL is invalid: '{stored_url}'."
            )
        if expected_port is None:
            raise CognitoConfigurationError(
                f"Bloom runtime {label} URL is invalid: '{expected_url}'."
            )

        if stored_port != expected_port:
            raise CognitoConfigurationError(
                f"Cognito {label} URL port mismatch for pool '{pool_id}' client '{client_id}': "
                f"stored '{stored_url}' uses port {stored_port}, but Bloom is running with '{expected_url}' "
                f"(port {expected_port}). Update the Cognito app callback/logout URLs to this Bloom port."
            )

    @classmethod
    def from_settings(
        cls,
        auth_settings,
        *,
        expected_callback_url: str = "",
        expected_logout_url: str = "",
        required_client_name: str = "",
    ) -> "CognitoAuth":
        """Create CognitoAuth from BloomSettings.auth instance.
        
        Args:
            auth_settings: AuthSettings instance from bloom_lims.config
        """
        pool_id = (auth_settings.cognito_user_pool_id or "").strip()
        region = (auth_settings.cognito_region or "").strip()
        client_id = (auth_settings.cognito_client_id or "").strip()
        client_secret = (auth_settings.cognito_client_secret or "").strip()
        client_name = (required_client_name or "bloom").strip()
        domain = (auth_settings.cognito_domain or "").strip()
        redirect_uri = (auth_settings.cognito_redirect_uri or "").strip()
        logout_uri = (auth_settings.cognito_logout_redirect_uri or redirect_uri).strip()
        expected_callback_url = (expected_callback_url or "").strip()
        expected_logout_url = (expected_logout_url or "").strip()

        if required_client_name:
            if client_name.lower() != required_client_name.lower():
                raise CognitoConfigurationError(
                    f"Cognito app client name mismatch for pool '{pool_id or '(unset)'}' client "
                    f"'{client_id or '(unset)'}': selected '{client_name}', expected '{required_client_name}'. "
                    "Update Bloom YAML config to use the bloom app client."
                )

        if expected_callback_url:
            cls._validate_runtime_uri_port_match(
                label="callback",
                stored_url=redirect_uri,
                expected_url=expected_callback_url,
                pool_id=pool_id or "(unset)",
                client_id=client_id or "(unset)",
            )
        if expected_logout_url:
            cls._validate_runtime_uri_port_match(
                label="logout",
                stored_url=logout_uri,
                expected_url=expected_logout_url,
                pool_id=pool_id or "(unset)",
                client_id=client_id or "(unset)",
            )

        required = {
            "cognito_user_pool_id": pool_id,
            "cognito_client_id": client_id,
            "cognito_region": region,
            "cognito_domain": domain,
            "cognito_redirect_uri": redirect_uri,
        }
        missing = [name for name, value in required.items() if not value]
        if missing:
            raise CognitoConfigurationError(
                f"Missing Cognito configuration in YAML config: {', '.join(sorted(missing))}"
            )

        config = CognitoConfig(
            user_pool_id=pool_id,
            client_id=client_id,
            client_secret=client_secret,
            region=region,
            domain=domain,
            redirect_uri=redirect_uri,
            logout_redirect_uri=logout_uri,
            scopes=auth_settings.cognito_scopes or ["openid", "email", "profile"],
        )
        return cls(config)

    def exchange_authorization_code(self, code: str) -> dict:
        """Exchange an OAuth2 authorization code for Cognito tokens."""
        payload = {
            "grant_type": "authorization_code",
            "client_id": self.config.client_id,
            "code": code,
            "redirect_uri": self.config.redirect_uri,
        }
        if self.config.client_secret:
            payload["client_secret"] = self.config.client_secret

        try:
            response = httpx.post(
                self.config.token_url,
                data=payload,
                headers={"Content-Type": "application/x-www-form-urlencoded"},
                timeout=15.0,
            )
        except Exception as exc:  # noqa: BLE001
            raise CognitoTokenError(f"Token exchange request failed: {exc}") from exc

        if response.status_code != 200:
            detail = (response.text or "").strip()[:500]
            raise CognitoTokenError(
                f"Token exchange failed ({response.status_code}): {detail}"
            )

        try:
            token_payload = response.json()
        except ValueError as exc:
            raise CognitoTokenError("Token exchange returned non-JSON response") from exc

        if not token_payload.get("id_token") and not token_payload.get("access_token"):
            raise CognitoTokenError("Token exchange returned no usable tokens")

        return token_payload

    def validate_token(self, token: str) -> dict:
        try:
            signing_key = self._jwks_client.get_signing_key_from_jwt(token)
            return jwt.decode(
                token,
                signing_key.key,
                algorithms=["RS256"],
                audience=self.config.client_id,
                issuer=self.config.issuer,
            )
        except Exception as exc:  # noqa: BLE001
            raise CognitoTokenError(str(exc)) from exc


@lru_cache(maxsize=1)
def get_cognito_auth() -> CognitoAuth:
    """Create (and cache) a CognitoAuth instance from YAML config.
    """
    from bloom_lims.config import get_settings

    settings = get_settings()
    return CognitoAuth.from_settings(settings.auth)



"""
BLOOM LIMS Configuration Management.

Configuration precedence (highest to lowest):
1. Environment variables (BLOOM_* prefix)
2. User config file (~/.config/bloom-<deployment>/bloom-config-<deployment>.yaml)
3. Template defaults
"""

import importlib.metadata
import logging
import os
import secrets
import string
import re
import tempfile
from functools import lru_cache
from importlib import resources as importlib_resources
from pathlib import Path
from typing import Any, Dict, List, Optional

from packaging.version import Version
from pydantic import BaseModel, Field, field_validator, model_validator
from pydantic_settings import BaseSettings, SettingsConfigDict
from pydantic_settings.sources import (
    PydanticBaseSettingsSource,
    YamlConfigSettingsSource,
)

from bloom_lims.domain_access import APPROVED_WEB_DOMAIN_SUFFIXES

logger = logging.getLogger(__name__)

DEFAULT_BLOOM_WEB_PORT = 8912
DEFAULT_BLOOM_TAPDB_LOCAL_PG_PORT = 5566

TEMPLATE_CONFIG_FILE = (
    Path(__file__).resolve().parent / "etc" / "bloom-config-template.yaml"
)


def _sanitize_deployment_code(value: str) -> str:
    cleaned = str(value or "").strip()
    cleaned = re.sub(r"[^A-Za-z0-9-]+", "-", cleaned)
    cleaned = cleaned.strip("-")
    return cleaned or "local"


def _resolve_deployment_code() -> str:
    return _sanitize_deployment_code(
        os.environ.get("BLOOM_DEPLOYMENT_CODE")
        or os.environ.get("DEPLOYMENT_CODE")
        or os.environ.get("LSMC_DEPLOYMENT_CODE")
        or "local"
    )


def _user_config_dir() -> Path:
    xdg_config_home = Path(os.environ.get("XDG_CONFIG_HOME") or (Path.home() / ".config"))
    return xdg_config_home / f"bloom-{_resolve_deployment_code()}"


def _user_config_file() -> Path:
    deployment = _resolve_deployment_code()
    return _user_config_dir() / f"bloom-config-{deployment}.yaml"


def _deployment_scoped_tapdb_config_path(client_id: str, namespace: str) -> str:
    xdg_config_home = Path(os.environ.get("XDG_CONFIG_HOME") or (Path.home() / ".config"))
    scoped_namespace = f"{namespace}-{_resolve_deployment_code()}"
    return str(xdg_config_home / "tapdb" / client_id / scoped_namespace / "tapdb-config.yaml")


def _validate_optional_https_url(value: str, *, field_name: str) -> str:
    normalized = str(value or "").strip()
    if not normalized:
        return ""
    if not normalized.startswith("https://"):
        raise ValueError(f"{field_name} must use an absolute https:// URL")
    return normalized.rstrip("/")


def _deep_merge(base: Dict[str, Any], override: Dict[str, Any]) -> Dict[str, Any]:
    """Deep merge two dictionaries. Override values take precedence."""
    result = base.copy()
    for key, value in override.items():
        if key in result and isinstance(result[key], dict) and isinstance(value, dict):
            result[key] = _deep_merge(result[key], value)
        else:
            result[key] = value
    return result


def _load_template_text() -> str:
    try:
        template_text = (
            importlib_resources.files("bloom_lims")
            .joinpath("etc/bloom-config-template.yaml")
            .read_text(encoding="utf-8")
        )
    except Exception as exc:
        logger.debug("Failed to load packaged template config text: %s", exc)
        return ""

    return template_text.replace(
        "__BLOOM_WEB_PORT__", str(DEFAULT_BLOOM_WEB_PORT)
    ).replace(
        "__BLOOM_TAPDB_LOCAL_PG_PORT__", str(DEFAULT_BLOOM_TAPDB_LOCAL_PG_PORT)
    )


def _load_template_config() -> Dict[str, Any]:
    """Load the packaged template configuration."""
    try:
        import yaml
    except ImportError:
        return {}

    template_text = _load_template_text()
    if not template_text:
        return {}

    try:
        loaded = yaml.safe_load(template_text) or {}
    except Exception as exc:
        logger.debug("Failed to parse packaged template config: %s", exc)
        return {}

    return loaded if isinstance(loaded, dict) else {}


def _rendered_template_config_file() -> Path:
    rendered_path = (
        Path(tempfile.gettempdir())
        / f"bloom-config-template-{_resolve_deployment_code()}.yaml"
    )
    template_text = _load_template_text()
    if template_text:
        try:
            if (
                not rendered_path.exists()
                or rendered_path.read_text(encoding="utf-8") != template_text
            ):
                rendered_path.write_text(template_text, encoding="utf-8")
        except Exception as exc:
            logger.debug("Failed to materialize rendered template config: %s", exc)
    return rendered_path


def _load_yaml_config() -> Dict[str, Any]:
    """Load and merge template + user YAML config files."""
    try:
        import yaml
    except ImportError:
        return {}

    config: Dict[str, Any] = {}

    template_config = _load_template_config()
    if template_config:
        config = template_config

    _ensure_user_config_dir()

    user_config_file = _user_config_file()
    if user_config_file.exists():
        try:
            with open(user_config_file, encoding="utf-8") as f:
                user_config = yaml.safe_load(f) or {}
                config = _deep_merge(config, user_config)
        except Exception as exc:
            logger.debug(
                "Failed to load user config %s: %s",
                user_config_file,
                exc,
            )

    return config


def _ensure_user_config_dir() -> None:
    """Ensure the user config directory exists when the process can create it."""
    user_config_dir = _user_config_dir()
    if user_config_dir.exists():
        return

    try:
        user_config_dir.mkdir(parents=True, exist_ok=True)
    except Exception as exc:
        logger.debug(
            "Failed to create user config directory %s: %s",
            user_config_dir,
            exc,
        )


def _yaml_config_files() -> list[Path]:
    _ensure_user_config_dir()
    return [_rendered_template_config_file(), _user_config_file()]


class TapDBSettings(BaseModel):
    """TapDB runtime targeting configuration."""

    env: str = Field(
        default="dev", description="TapDB target environment: dev/test/prod"
    )
    client_id: str = Field(
        default="bloom",
        description="TapDB client namespace key for Bloom runtime config",
    )
    database_name: str = Field(
        default="bloom",
        description="TapDB database namespace key for Bloom runtime config",
    )
    strict_namespace: bool = Field(
        default=True,
        description="Enable strict namespace mode for namespaced TapDB config",
    )
    config_path: str = Field(
        default="",
        description="Optional explicit TapDB config path override",
    )
    local_pg_port: int = Field(
        default=DEFAULT_BLOOM_TAPDB_LOCAL_PG_PORT,
        description="Default local PostgreSQL port for TapDB dev/test runtime",
    )
    min_version: str = Field(
        default="3.0.6", description="Minimum supported daylily-tapdb"
    )
    max_version_exclusive: str = Field(
        default="4.0.0",
        description="Exclusive upper bound for daylily-tapdb",
    )

    @field_validator("env")
    @classmethod
    def validate_env(cls, value: str) -> str:
        env = value.strip().lower()
        if env not in {"dev", "test", "prod"}:
            raise ValueError("tapdb.env must be one of: dev, test, prod")
        return env

    @field_validator("client_id")
    @classmethod
    def validate_client_id(cls, value: str) -> str:
        cleaned = value.strip()
        if not cleaned:
            raise ValueError("tapdb.client_id cannot be empty")
        return cleaned

    @field_validator("database_name")
    @classmethod
    def validate_database_name(cls, value: str) -> str:
        cleaned = value.strip()
        if not cleaned:
            raise ValueError("tapdb.database_name cannot be empty")
        return cleaned

    @field_validator("local_pg_port")
    @classmethod
    def validate_local_pg_port(cls, value: int) -> int:
        if value < 1 or value > 65535:
            raise ValueError("tapdb.local_pg_port must be between 1 and 65535")
        return value


class TapDBRuntimeContext(BaseModel):
    """Resolved runtime context to pass to daylily-tapdb."""

    env: str
    client_id: str
    database_name: str
    strict_namespace: bool = True
    config_path: str = ""
    aws_profile: str = "lsmc"
    aws_region: str = "us-west-2"


class StorageSettings(BaseModel):
    """File storage configuration."""

    upload_dir: str = Field(
        default="/var/lib/bloom/uploads", description="Upload directory"
    )
    temp_dir: str = Field(
        default_factory=lambda: str(Path(tempfile.gettempdir()) / "bloom"),
        description="Temporary file directory",
    )
    max_file_size_mb: int = Field(default=100, description="Max upload file size (MB)")
    allowed_extensions: List[str] = Field(
        default=["pdf", "csv", "xlsx", "txt", "json", "png", "jpg", "jpeg"],
        description="Allowed file extensions",
    )

    s3_bucket: str = Field(default="", description="S3 bucket name")
    s3_prefix: str = Field(default="bloom-files/", description="S3 key prefix")
    s3_region: str = Field(default="us-east-1", description="S3 region")
    s3_endpoint: str = Field(default="", description="S3-compatible endpoint URL")

    @property
    def max_file_size_bytes(self) -> int:
        return self.max_file_size_mb * 1024 * 1024


def _get_default_api_version() -> str:
    try:
        from bloom_lims._version import get_version

        return get_version()
    except ImportError:
        return "0.10.7"


class APISettings(BaseModel):
    """API configuration."""

    title: str = Field(default="BLOOM LIMS API", description="API title")
    version: str = Field(
        default_factory=_get_default_api_version, description="API version"
    )
    prefix: str = Field(default="/api/v1", description="API URL prefix")

    pagination_default_size: int = Field(default=50, description="Default page size")
    pagination_max_size: int = Field(default=1000, description="Maximum page size")

    rate_limit_enabled: bool = Field(default=True, description="Enable rate limiting")
    rate_limit_requests: int = Field(default=100, description="Requests per window")
    rate_limit_window_seconds: int = Field(default=60, description="Rate limit window")

    request_timeout_seconds: int = Field(default=30, description="Request timeout")
    long_running_timeout_seconds: int = Field(
        default=300, description="Long-running op timeout"
    )

    cors_origins: List[str] = Field(
        default_factory=lambda: [
            f"https://{item}" for item in APPROVED_WEB_DOMAIN_SUFFIXES
        ],
        description="Allowed CORS origins",
    )
    cors_allow_credentials: bool = Field(default=True, description="Allow credentials")


class AWSSettings(BaseModel):
    """AWS configuration."""

    profile: str = Field(default="lsmc", description="AWS profile name")
    region: str = Field(default="us-west-2", description="Default AWS region")


class UISettings(BaseModel):
    """UI metadata configuration."""

    support_email: str = Field(
        default="support@dyly.bio",
        description="Support contact email shown in the GUI footer/help page",
    )
    github_repo_url: str = Field(
        default="https://github.com/Daylily-Informatics/bloom",
        description="Repository URL shown in the GUI footer/help page",
    )


class DeploymentSettings(BaseModel):
    """Deployment-specific GUI chrome."""

    name: str = Field(default="", description="Deployment label")
    color: str = Field(default="#0f766e", description="Deployment banner color")
    is_production: bool = Field(
        default=False, description="Hide deployment banner in production"
    )


class AuthSettings(BaseModel):
    """Authentication configuration."""

    cognito_user_pool_id: str = Field(default="", description="Cognito user pool ID")
    cognito_client_id: str = Field(default="", description="Cognito app client ID")
    cognito_client_secret: str = Field(
        default="", description="Cognito app client secret"
    )
    cognito_region: str = Field(default="", description="AWS region for Cognito")
    cognito_domain: str = Field(default="", description="Cognito hosted UI domain")
    cognito_redirect_uri: str = Field(default="", description="Cognito redirect URI")
    cognito_logout_redirect_uri: str = Field(
        default="", description="Cognito logout redirect URI"
    )
    cognito_scopes: List[str] = Field(
        default_factory=lambda: ["openid", "email", "profile"],
        description="Cognito OAuth scopes",
    )
    cognito_allowed_domains: List[str] = Field(
        default_factory=lambda: list(APPROVED_WEB_DOMAIN_SUFFIXES),
        description="Allowed email domains",
    )

    jwt_secret: str = Field(default="", description="JWT secret key")
    jwt_algorithm: str = Field(default="HS256", description="JWT algorithm")
    jwt_expiry_hours: int = Field(default=24, description="JWT token expiry (hours)")

    session_timeout_minutes: int = Field(default=30, description="Session timeout")
    max_sessions_per_user: int = Field(default=5, description="Max concurrent sessions")


class AtlasSettings(BaseModel):
    """Atlas integration settings."""

    base_url: str = Field(default="", description="Atlas API base URL")
    token: str = Field(default="", description="Atlas API bearer token")
    timeout_seconds: int = Field(default=10, description="Atlas API timeout seconds")
    cache_ttl_seconds: int = Field(
        default=300, description="Atlas response cache TTL in seconds"
    )
    verify_ssl: bool = Field(default=True, description="Verify Atlas TLS certificates")
    organization_id: str = Field(
        default="",
        description="Atlas organization/tenant UUID used for outbound Bloom events",
    )
    events_enabled: bool = Field(
        default=False,
        description="Enable outbound Bloom->Atlas webhook event delivery",
    )
    events_path: str = Field(
        default="/api/integrations/bloom/v1/events",
        description="Atlas webhook path for Bloom events",
    )
    webhook_secret: str = Field(
        default="",
        description="Shared HMAC secret for Bloom webhook signatures",
    )
    events_timeout_seconds: int = Field(
        default=10,
        description="Timeout (seconds) for outbound Bloom webhook delivery",
    )
    events_max_retries: int = Field(
        default=2,
        description="Maximum retry count for outbound Bloom webhook delivery",
    )
    status_events_timeout_seconds: int = Field(
        default=10,
        description="Timeout (seconds) for Bloom -> Atlas test-order status event calls",
    )
    status_events_max_retries: int = Field(
        default=5,
        description="Maximum retry count for Bloom -> Atlas status event calls",
    )
    status_events_backoff_base_seconds: float = Field(
        default=0.5,
        description="Base backoff delay in seconds for Bloom -> Atlas status event retries",
    )

    @field_validator("base_url")
    @classmethod
    def validate_base_url(cls, value: str) -> str:
        return _validate_optional_https_url(value, field_name="atlas.base_url")


class DeweySettings(BaseModel):
    """Dewey integration settings."""

    enabled: bool = Field(
        default=False, description="Enable Bloom -> Dewey artifact registration"
    )
    base_url: str = Field(default="", description="Dewey API base URL")
    token: str = Field(default="", description="Dewey API bearer token")
    timeout_seconds: int = Field(default=10, description="Dewey API timeout seconds")
    verify_ssl: bool = Field(default=True, description="Verify Dewey TLS certificates")

    @field_validator("base_url")
    @classmethod
    def validate_base_url(cls, value: str) -> str:
        return _validate_optional_https_url(value, field_name="dewey.base_url")

    @model_validator(mode="after")
    def validate_enabled_contract(self) -> "DeweySettings":
        if self.enabled:
            if not str(self.base_url or "").strip():
                raise ValueError("dewey.base_url is required when dewey.enabled=true")
            if not str(self.token or "").strip():
                raise ValueError("dewey.token is required when dewey.enabled=true")
        return self


class ZebraDaySettings(BaseModel):
    """zebra_day integration settings."""

    base_url: str = Field(default="", description="zebra_day API base URL")
    token: str = Field(default="", description="zebra_day internal API bearer token")
    timeout_seconds: int = Field(default=10, description="zebra_day API timeout seconds")
    verify_ssl: bool = Field(
        default=True, description="Verify zebra_day TLS certificates"
    )

    @field_validator("base_url")
    @classmethod
    def validate_base_url(cls, value: str) -> str:
        normalized = str(value or "").strip()
        if not normalized:
            return ""
        if not normalized.startswith(("https://", "http://")):
            raise ValueError("zebra_day.base_url must use an absolute http:// or https:// URL")
        return normalized.rstrip("/")


class LoggingSettings(BaseModel):
    """Logging configuration."""

    level: str = Field(default="INFO", description="Log level")
    format: str = Field(
        default="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
        description="Log format",
    )
    json_format: bool = Field(default=False, description="Use JSON log format")
    log_file: str = Field(default="", description="Log file path (empty for stdout)")
    max_file_size_mb: int = Field(default=10, description="Max log file size")
    backup_count: int = Field(default=5, description="Number of backup files")


class FeatureFlags(BaseModel):
    """Feature flags for enabling/disabling features."""

    enable_audit_logging: bool = Field(default=True, description="Enable audit trail")
    enable_workflow_notifications: bool = Field(
        default=True, description="Workflow notifications"
    )
    enable_file_versioning: bool = Field(
        default=True, description="Enable file versioning"
    )
    enable_advanced_search: bool = Field(
        default=True, description="Advanced search features"
    )
    enable_api_caching: bool = Field(
        default=False, description="Enable API response caching"
    )
    maintenance_mode: bool = Field(default=False, description="Enable maintenance mode")


class CacheSettings(BaseModel):
    """Cache configuration for application caching layer."""

    enabled: bool = Field(default=True, description="Enable caching")
    backend: str = Field(
        default="memory", description="Cache backend: memory, redis, memcached"
    )
    max_size: int = Field(default=5000, description="Maximum cache entries")
    default_ttl: int = Field(default=300, description="Default TTL in seconds")

    template_ttl: int = Field(default=3600, description="Template cache TTL")
    instance_ttl: int = Field(default=300, description="Instance cache TTL")
    lineage_ttl: int = Field(default=300, description="Lineage cache TTL")
    query_ttl: int = Field(default=60, description="Query result cache TTL")

    template_prefix: str = Field(
        default="tmpl:", description="Template cache key prefix"
    )
    instance_prefix: str = Field(
        default="inst:", description="Instance cache key prefix"
    )
    query_prefix: str = Field(default="qry:", description="Query cache key prefix")

    redis_host: str = Field(default="localhost", description="Redis server host")
    redis_port: int = Field(default=6379, description="Redis server port")
    redis_db: int = Field(default=0, description="Redis database number")
    redis_password: Optional[str] = Field(default=None, description="Redis password")
    redis_ssl: bool = Field(default=False, description="Enable Redis SSL/TLS")
    redis_cluster: bool = Field(default=False, description="Enable Redis Cluster mode")

    memcached_servers: List[str] = Field(
        default=["localhost:11211"],
        description="Memcached server addresses",
    )


class BusinessConstants(BaseModel):
    """Business logic constants extracted from code."""

    default_version: str = Field(default="1.0", description="Default object version")
    wildcard_version: str = Field(default="*", description="Wildcard version string")

    status_created: str = Field(default="created", description="Created status")
    status_in_progress: str = Field(
        default="in_progress", description="In progress status"
    )
    status_complete: str = Field(default="complete", description="Complete status")
    status_failed: str = Field(default="failed", description="Failed status")
    status_abandoned: str = Field(default="abandoned", description="Abandoned status")
    status_active: str = Field(default="active", description="Active status")

    workflow_state_active: str = Field(
        default="active", description="Active workflow state"
    )
    workflow_state_paused: str = Field(
        default="paused", description="Paused workflow state"
    )
    workflow_state_completed: str = Field(
        default="completed", description="Completed workflow state"
    )

    template_suffix: str = Field(
        default="_template", description="Template table suffix"
    )
    instance_suffix: str = Field(
        default="_instance", description="Instance table suffix"
    )
    lineage_suffix: str = Field(
        default="_instance_lineage", description="Lineage table suffix"
    )

    singleton_true_values: List[str] = Field(
        default=["1", "true", "True", "yes", "Yes"],
        description="Values that indicate singleton=True",
    )
    singleton_false_values: List[str] = Field(
        default=["0", "false", "False", "no", "No", ""],
        description="Values that indicate singleton=False",
    )

    max_lineage_depth: int = Field(
        default=10, description="Max depth for lineage queries"
    )
    max_children_per_query: int = Field(
        default=10000, description="Max children in recursive queries"
    )

    default_label_style: str = Field(
        default="tube_2inX1in", description="Default label style"
    )
    default_lab_code: str = Field(default="BLOOM", description="Default lab code")

    default_timezone: str = Field(default="UTC", description="Default timezone")

    config_base_path: str = Field(
        default="config", description="Base path for template configurations"
    )

    action_max_executions_unlimited: str = Field(
        default="-1", description="Unlimited action executions"
    )
    action_enabled_true: str = Field(default="1", description="Action enabled value")
    action_enabled_false: str = Field(default="0", description="Action disabled value")


class BloomSettings(BaseSettings):
    """Main settings class for BLOOM LIMS."""

    model_config = SettingsConfigDict(
        env_prefix="BLOOM_",
        env_nested_delimiter="__",
        case_sensitive=False,
        extra="ignore",
    )

    app_name: str = Field(default="BLOOM LIMS", description="Application name")
    environment: str = Field(default="development", description="Environment name")
    debug: bool = Field(default=False, description="Debug mode")

    tapdb: TapDBSettings = Field(default_factory=TapDBSettings)
    storage: StorageSettings = Field(default_factory=StorageSettings)
    api: APISettings = Field(default_factory=APISettings)
    auth: AuthSettings = Field(default_factory=AuthSettings)
    atlas: AtlasSettings = Field(default_factory=AtlasSettings)
    dewey: DeweySettings = Field(default_factory=DeweySettings)
    zebra_day: ZebraDaySettings = Field(default_factory=ZebraDaySettings)
    aws: AWSSettings = Field(default_factory=AWSSettings)
    ui: UISettings = Field(default_factory=UISettings)
    deployment: DeploymentSettings = Field(default_factory=DeploymentSettings)
    logging: LoggingSettings = Field(default_factory=LoggingSettings)
    features: FeatureFlags = Field(default_factory=FeatureFlags)
    cache: CacheSettings = Field(default_factory=CacheSettings)
    constants: BusinessConstants = Field(default_factory=BusinessConstants)

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: type[BaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> tuple[PydanticBaseSettingsSource, ...]:
        return (
            init_settings,
            env_settings,
            dotenv_settings,
            file_secret_settings,
            YamlConfigSettingsSource(
                settings_cls,
                yaml_file=_yaml_config_files(),
                yaml_file_encoding="utf-8",
            ),
        )

    @model_validator(mode="after")
    def validate_dewey_settings(self) -> "BloomSettings":
        if self.dewey.enabled:
            if not str(self.dewey.base_url or "").strip():
                raise ValueError("dewey.base_url is required when dewey.enabled=true")
            if not str(self.dewey.token or "").strip():
                raise ValueError("dewey.token is required when dewey.enabled=true")
        return self

    @field_validator("environment")
    @classmethod
    def validate_environment(cls, value: str) -> str:
        env = value.lower()
        valid_envs = ["development", "staging", "production", "testing"]
        if env not in valid_envs:
            raise ValueError(f"Environment must be one of: {', '.join(valid_envs)}")
        return env

    @property
    def is_production(self) -> bool:
        return self.environment == "production"

    @property
    def is_development(self) -> bool:
        return self.environment == "development"

    # Compatibility aliases for explicit tapdb context naming.
    @property
    def tapdb_env(self) -> str:
        return self.tapdb.env

    @property
    def tapdb_database_name(self) -> str:
        return self.tapdb.database_name

    @property
    def tapdb_config_path(self) -> str:
        return self.tapdb.config_path


@lru_cache()
def get_settings() -> BloomSettings:
    return BloomSettings()


def get_tapdb_runtime_context(
    settings: Optional[BloomSettings] = None,
) -> TapDBRuntimeContext:
    """Resolve active TapDB runtime context from Bloom config."""
    settings = settings or get_settings()
    return TapDBRuntimeContext(
        env=settings.tapdb.env.strip().lower(),
        client_id=settings.tapdb.client_id.strip(),
        database_name=settings.tapdb.database_name.strip(),
        strict_namespace=settings.tapdb.strict_namespace,
        config_path=(
            settings.tapdb.config_path
            or _deployment_scoped_tapdb_config_path(
                settings.tapdb.client_id.strip(),
                settings.tapdb.database_name.strip(),
            )
        ).strip(),
        aws_profile=(settings.aws.profile or "lsmc").strip(),
        aws_region=(settings.aws.region or "us-west-2").strip(),
    )


def apply_runtime_environment(
    settings: Optional[BloomSettings] = None,
) -> TapDBRuntimeContext:
    """Return normalized Bloom runtime context without mutating TAPDB env."""
    settings = settings or get_settings()
    ctx = get_tapdb_runtime_context(settings=settings)
    return ctx


def get_tapdb_db_config(
    env_name: Optional[str] = None,
    database_name: Optional[str] = None,
    config_path: Optional[str] = None,
) -> Dict[str, str]:
    """Resolve active DB config via daylily-tapdb config loader."""
    settings = get_settings()
    ctx = apply_runtime_environment(settings)

    target_env = (env_name or ctx.env).strip().lower()

    from daylily_tapdb.cli.db_config import get_db_config_for_env

    return get_db_config_for_env(
        target_env,
        config_path=(config_path or ctx.config_path or None),
        client_id=ctx.client_id,
        database_name=(database_name or ctx.database_name),
    )


def assert_tapdb_version(
    min_version: Optional[str] = None,
    max_version_exclusive: Optional[str] = None,
) -> str:
    """Assert installed daylily-tapdb version is within the supported range."""
    settings = get_settings()
    min_v = Version(min_version or settings.tapdb.min_version)
    max_v = Version(max_version_exclusive or settings.tapdb.max_version_exclusive)

    try:
        installed = Version(importlib.metadata.version("daylily-tapdb"))
    except importlib.metadata.PackageNotFoundError as exc:
        raise RuntimeError("daylily-tapdb is not installed") from exc

    if installed < min_v or installed >= max_v:
        raise RuntimeError(
            f"Unsupported daylily-tapdb version {installed}; expected >= {min_v} and < {max_v}"
        )

    return str(installed)


def generate_example_webhook_secret(length: int = 20) -> str:
    alphabet = string.ascii_letters + string.digits
    return "".join(secrets.choice(alphabet) for _ in range(length))


def atlas_webhook_secret_warning(
    settings: "BloomSettings | None" = None,
    *,
    suggested_secret: str | None = None,
) -> str | None:
    active_settings = settings or get_settings()
    if str(active_settings.atlas.webhook_secret or "").strip():
        return None
    example = suggested_secret or generate_example_webhook_secret()
    return (
        "Atlas webhook signature secret is not configured "
        "(atlas.webhook_secret is empty). Bloom can still start, but signed Atlas webhook "
        "delivery and verification will not be safe or interoperable. Set a 20-character "
        f"alphanumeric secret such as: {example}"
    )


def validate_settings() -> List[str]:
    """Validate settings and return list of warnings/issues."""
    warnings: List[str] = []
    settings = get_settings()

    try:
        assert_tapdb_version()
    except Exception as exc:
        warnings.append(f"TapDB version check failed: {exc}")

    if not settings.auth.cognito_user_pool_id:
        warnings.append(
            "Cognito configuration is missing (auth.cognito_user_pool_id). "
            "Set it in YAML config or override it with BLOOM_AUTH__COGNITO_USER_POOL_ID."
        )

    if not settings.auth.jwt_secret and settings.is_production:
        warnings.append("JWT secret is not set in production")

    upload_path = Path(settings.storage.upload_dir)
    if not upload_path.exists():
        warnings.append(
            f"Upload directory does not exist: {settings.storage.upload_dir}"
        )

    if settings.storage.s3_bucket and not os.environ.get("AWS_ACCESS_KEY_ID"):
        warnings.append("S3 bucket configured but AWS_ACCESS_KEY_ID not set")

    atlas_secret_warning = atlas_webhook_secret_warning(settings)
    if atlas_secret_warning:
        warnings.append(atlas_secret_warning)

    return warnings


def validate_config_content(content: str) -> List[str]:
    """Validate YAML config content without consulting runtime environment."""
    try:
        import yaml
    except ImportError:
        return ["PyYAML is required to validate configuration"]

    try:
        parsed = yaml.safe_load(content) or {}
    except yaml.YAMLError as exc:
        return [f"YAML parse error: {exc}"]

    if not isinstance(parsed, dict):
        return ["Root YAML object must be a mapping"]

    merged = _deep_merge(_load_template_config(), parsed)

    try:
        BloomSettings.model_validate(merged)
    except Exception as exc:
        details = getattr(exc, "errors", None)
        if callable(details):
            messages: List[str] = []
            for item in details():
                location = ".".join(str(part) for part in item.get("loc", ()))
                message = item.get("msg", str(exc))
                if location:
                    messages.append(f"{location}: {message}")
                else:
                    messages.append(message)
            if messages:
                return messages
        return [str(exc)]

    return []


# Legacy compatibility
def get_database_url() -> str:
    """Build runtime database URL from TapDB resolved configuration."""
    cfg = get_tapdb_db_config()
    auth = cfg["user"]
    if cfg.get("password"):
        auth = f"{cfg['user']}:{cfg['password']}"

    url = f"postgresql://{auth}@{cfg['host']}:{cfg['port']}/{cfg['database']}"
    if cfg.get("engine_type") == "aurora":
        return f"{url}?sslmode=verify-full"
    return url


def get_cognito_config() -> tuple[str, str, str]:
    settings = get_settings()
    return (
        settings.auth.cognito_region,
        settings.auth.cognito_user_pool_id,
        settings.auth.cognito_client_id,
    )


def get_user_config_path() -> Path:
    return _user_config_file()


def get_template_config_path() -> Path:
    return TEMPLATE_CONFIG_FILE


def ensure_user_config_exists() -> Path:
    """Ensure user config directory and file exist."""
    user_config_dir = _user_config_dir()
    user_config_file = _user_config_file()
    if not user_config_dir.exists():
        user_config_dir.mkdir(parents=True, exist_ok=True)

    if not user_config_file.exists():
        template_text = _load_template_text()
        if template_text:
            user_config_file.write_text(template_text, encoding="utf-8")

    return user_config_file



"""BLOOM LIMS CLI — built on cli-core-yo."""

from __future__ import annotations

import os
import sys
from pathlib import Path

from cli_core_yo.app import create_app as _create_app
from cli_core_yo.app import run
from cli_core_yo.spec import CliSpec, ConfigSpec, PluginSpec, XdgSpec

from bloom_lims.config import (
    _load_template_text,
    _resolve_deployment_code,
    apply_runtime_environment,
    assert_tapdb_version,
    get_settings,
    get_tapdb_db_config,
    validate_config_content,
)

PROJECT_ROOT = Path(__file__).resolve().parents[2]


def _validate_bloom_config(content: str) -> list[str]:
    """Validate Bloom config file contents."""
    return validate_config_content(content)


def _bloom_info_hook() -> list[tuple[str, str]]:
    """Extra rows for the built-in ``info`` command."""
    settings = get_settings()
    ctx = apply_runtime_environment(settings)

    rows: list[tuple[str, str]] = [
        ("Project Root", str(PROJECT_ROOT)),
        ("Environment", settings.environment),
        ("TapDB Env", ctx.env),
        ("TapDB Namespace", ctx.database_name),
        ("TapDB Config", ctx.config_path or "(derived)"),
        ("AWS Profile", os.environ.get("AWS_PROFILE", ctx.aws_profile)),
        ("AWS Region", os.environ.get("AWS_REGION", ctx.aws_region)),
        ("Conda Env", os.environ.get("CONDA_DEFAULT_ENV", "(not set)")),
    ]

    try:
        db_cfg = get_tapdb_db_config()
        rows.append(
            (
                "TapDB Target",
                f"{db_cfg['host']}:{db_cfg['port']}/{db_cfg['database']}",
            )
        )
    except Exception as exc:
        rows.append(("TapDB Target", f"(unresolved: {exc})"))

    try:
        rows.append(("daylily-tapdb", assert_tapdb_version()))
    except Exception as exc:
        rows.append(("daylily-tapdb", f"invalid ({exc})"))

    from bloom_lims.cli.server import server_status_label

    rows.append(("Dev Server", server_status_label()))
    return rows


spec = CliSpec(
    prog_name="bloom",
    app_display_name="BLOOM LIMS",
    dist_name="bloom_lims",
    root_help="BLOOM LIMS — Development CLI for the laboratory information management system.",
    xdg=XdgSpec(
        app_dir_name=f"bloom-{_resolve_deployment_code()}",
    ),
    config=ConfigSpec(
        primary_filename=f"bloom-config-{_resolve_deployment_code()}.yaml",
        template_bytes=_load_template_text().encode("utf-8"),
        validator=_validate_bloom_config,
    ),
    plugins=PluginSpec(
        explicit=[
            "bloom_lims.cli.server.register",
            "bloom_lims.cli.db.register",
            "bloom_lims.cli.test.register",
            "bloom_lims.cli.quality.register",
            "bloom_lims.cli.users.register",
            "bloom_lims.cli.integrations.register",
            "bloom_lims.cli.config_extra.register",
        ],
    ),
    info_hooks=[_bloom_info_hook],
)


def build_app():
    """Create a fresh Typer app for Bloom."""
    return _create_app(spec)


app = build_app()

_SKIP_CONDA_ENV_CHECK_FLAG = "--skip-conda-env-check"
_CONDA_ENV_CHECK_EXEMPT_COMMANDS = frozenset({"version", "info", "help"})


def _strip_skip_conda_env_check_flag(args: list[str]) -> tuple[list[str], bool]:
    filtered = [arg for arg in args if arg != _SKIP_CONDA_ENV_CHECK_FLAG]
    return filtered, len(filtered) != len(args)


def _command_requires_conda_env_check(args: list[str]) -> bool:
    if not args or "--help" in args or "-h" in args:
        return False
    for arg in args:
        if not arg or arg.startswith("-"):
            continue
        return arg not in _CONDA_ENV_CHECK_EXEMPT_COMMANDS
    return False


def _enforce_conda_env_contract(args: list[str]) -> None:
    if not _command_requires_conda_env_check(args):
        return
    active_env = os.environ.get("CONDA_DEFAULT_ENV", "").strip()
    if not active_env:
        raise SystemExit(
            "Bloom CLI requires an active deployment-scoped conda environment. "
            "Activate an env named like 'BLOOM-local2', or pass "
            "--skip-conda-env-check to override."
        )
    if "-" not in active_env:
        raise SystemExit(
            f"Bloom CLI requires a deployment-scoped conda environment name with '-'. "
            f"Current CONDA_DEFAULT_ENV='{active_env}'. Pass --skip-conda-env-check to override."
        )


def main() -> None:
    """Main CLI entry point."""
    args, skip_conda_env_check = _strip_skip_conda_env_check_flag(sys.argv[1:])
    if not skip_conda_env_check:
        _enforce_conda_env_contract(args)
    raise SystemExit(run(spec, args))


if __name__ == "__main__":
    main()



"""
BLOOM LIMS Test Configuration and Fixtures

This module provides pytest fixtures and utilities for testing BLOOM LIMS.

Usage:
    # In test files, fixtures are automatically available
    def test_something(bdb, bloom_obj):
        result = bloom_obj.get_by_euid("WF_ABC123_X")
        assert result is not None

Fixtures:
    - bdb: Database connection (session-scoped)
    - bloom_obj: BloomObj instance
    - test_template: A test template for creating instances
    - clean_test_data: Cleanup fixture for test data
"""

import logging
from unittest.mock import MagicMock, patch

import pytest

from tests.support.runtime import (
    ensure_local_tapdb_ready,
    ensure_test_runtime_environment,
    selected_items_need_local_tapdb,
)

# Configure test logging
logging.basicConfig(level=logging.WARNING)
logger = logging.getLogger(__name__)


# Environment setup for tests
def pytest_configure(config):
    """Configure pytest environment."""
    ensure_test_runtime_environment()
    config._bloom_needs_local_tapdb = False


@pytest.fixture(scope="session", autouse=True)
def _ensure_db_backed_integration_runtime(pytestconfig):
    """Provision local TapDB for DB-backed integration suites when needed."""
    if not getattr(pytestconfig, "_bloom_needs_local_tapdb", False):
        yield
        return

    try:
        ensure_local_tapdb_ready()
    except Exception as exc:
        pytest.exit(
            "BLOOM integration test bootstrap failed. "
            f"{exc} "
            "Retry after `source ./activate <deploy-name> && bloom db init`.",
            returncode=2,
        )

    yield


@pytest.fixture(scope="session")
def bdb():
    """
    Session-scoped database connection fixture.

    Creates a single database connection for all tests in the session.
    The connection is automatically closed after all tests complete.

    Yields:
        BLOOMdb3: Database connection instance
    """
    from bloom_lims.db import BLOOMdb3

    logger.info("Creating test database connection")
    db = BLOOMdb3()

    yield db

    logger.info("Closing test database connection")
    db.close()


@pytest.fixture(scope="function")
def bdb_function():
    """
    Function-scoped database connection fixture.

    Creates a new database connection for each test function.
    Use this when tests need isolated database state.

    Yields:
        BLOOMdb3: Database connection instance
    """
    from bloom_lims.db import BLOOMdb3

    db = BLOOMdb3()
    yield db
    db.close()


@pytest.fixture(scope="function")
def bloom_obj(bdb):
    """
    BloomObj instance fixture.

    Creates a BloomObj instance using the session-scoped database connection.

    Args:
        bdb: Database connection fixture

    Yields:
        BloomObj: BloomObj instance
    """
    from bloom_lims.bobjs import BloomObj

    return BloomObj(bdb)


@pytest.fixture(scope="function")
def bloom_container(bdb):
    """BloomContainer instance fixture."""
    from bloom_lims.bobjs import BloomContainer

    return BloomContainer(bdb)


@pytest.fixture(scope="function")
def bloom_workflow(bdb):
    """BloomWorkflow instance fixture."""
    from bloom_lims.bobjs import BloomWorkflow

    return BloomWorkflow(bdb)


@pytest.fixture(scope="function")
def bloom_content(bdb):
    """BloomContent instance fixture."""
    from bloom_lims.bobjs import BloomContent

    return BloomContent(bdb)


@pytest.fixture(scope="function")
def test_template(bdb, bloom_obj):
    """
    Get a test template for creating instances.

    Returns the first available generic_template.

    Args:
        bdb: Database connection fixture
        bloom_obj: BloomObj fixture

    Returns:
        Template object or None if no templates exist
    """
    templates = (
        bdb.session.query(bloom_obj.Base.classes.generic_template).limit(1).all()
    )
    return templates[0] if templates else None


@pytest.fixture(scope="function")
def mock_session():
    """
    Mock SQLAlchemy session for unit tests.

    Use this when you want to test without database access.

    Returns:
        MagicMock: Mocked session object
    """
    session = MagicMock()
    session.query.return_value.filter.return_value.first.return_value = None
    session.query.return_value.filter_by.return_value.first.return_value = None
    return session


@pytest.fixture(scope="function")
def clean_cache():
    """
    Clear the global cache before and after test.

    Use this when testing cache behavior.
    """
    from bloom_lims.core.cache import cache_clear

    cache_clear()
    yield
    cache_clear()


# API Test Fixtures
@pytest.fixture(scope="function")
def api_client():
    """
    FastAPI test client fixture.

    Creates a test client for the BLOOM LIMS API.

    Yields:
        TestClient: FastAPI test client
    """
    from bloom_lims.main import app
    from fastapi.testclient import TestClient

    return TestClient(app)


@pytest.fixture(scope="function")
def authenticated_api_client(api_client):
    """
    Authenticated API client fixture.

    Creates a test client with mock authentication headers.

    Args:
        api_client: Base API client fixture

    Yields:
        TestClient: Authenticated FastAPI test client
    """
    # Add mock auth header
    api_client.headers["X-API-Key"] = "test-api-key"
    return api_client


@pytest.fixture(scope="function")
def mock_api_auth():
    """
    Mock API authentication for testing.

    Patches the require_api_auth dependency to return a test user.

    Yields:
        APIUser: Mock user object
    """
    from bloom_lims.api.v1.dependencies import APIUser

    mock_user = APIUser(
        user_id="test-user-id",
        email="test@example.com",
        role="ADMIN",
    )

    with patch(
        "bloom_lims.api.v1.dependencies.require_api_auth", return_value=mock_user
    ):
        yield mock_user


# Test markers
def pytest_collection_modifyitems(config, items):
    """Add markers to tests based on their location."""
    config._bloom_needs_local_tapdb = selected_items_need_local_tapdb(items)
    for item in items:
        # Mark integration tests
        if "integration" in str(item.fspath):
            item.add_marker(pytest.mark.integration)

        # Mark slow tests
        if "slow" in item.name.lower():
            item.add_marker(pytest.mark.slow)

        # Mark API tests
        if "api" in str(item.fspath) or "api" in item.name.lower():
            item.add_marker(pytest.mark.api)



from __future__ import annotations

from urllib.parse import urlparse

from playwright.sync_api import Page, expect


def perform_login(page: Page, *, base_url: str, email: str, password: str) -> None:
    page.goto(f"{base_url}/login")
    expect(page.locator("body")).to_be_visible()
    page.locator("a:has-text('Sign in with AWS Cognito')").click()
    _complete_identity_provider_login(page, email=email, password=password)
    page.wait_for_url(f"{base_url}/**", timeout=30000)
    assert_authenticated_page(page)


def perform_logout(page: Page, *, base_url: str) -> None:
    page.goto(f"{base_url}/auth/logout")
    page.wait_for_load_state("domcontentloaded")


def assert_authenticated_page(page: Page) -> None:
    expect(page.locator("body")).to_be_visible()
    body = page.locator("body").inner_text().lower()
    for snippet in ("authentication failed", "not authorized", "not provisioned", "email domain not allowed"):
        assert snippet not in body, f"Unexpected auth failure content detected: {snippet}"


def _complete_identity_provider_login(page: Page, *, email: str, password: str) -> None:
    page.wait_for_load_state("domcontentloaded")
    try:
        if "accounts.google.com" in (urlparse(page.url).hostname or ""):
            _complete_google_login(page, email=email, password=password)
            return
        if page.locator("input[type='email']").first.is_visible(timeout=3000) and "accounts.google.com" in (urlparse(page.url).hostname or ""):
            _complete_google_login(page, email=email, password=password)
            return
    except Exception:
        pass
    _complete_cognito_login(page, email=email, password=password)


def _complete_cognito_login(page: Page, *, email: str, password: str) -> None:
    user_input = page.locator(
        "input[name='username']:visible, input[type='email']:visible"
    ).first
    pass_input = page.locator(
        "input[name='password']:visible, input[type='password']:visible"
    ).first
    submit_btn = page.locator(
        "input[name='signInSubmitButton']:visible, button[type='submit']:visible"
    ).first

    expect(user_input).to_be_visible(timeout=15000)
    user_input.fill(email)
    pass_input.fill(password)
    submit_btn.click()


def _complete_google_login(page: Page, *, email: str, password: str) -> None:
    email_input = page.locator("input[type='email']").first
    expect(email_input).to_be_visible(timeout=15000)
    email_input.fill(email)
    page.locator("#identifierNext, button:has-text('Next')").first.click()

    password_input = page.locator("input[type='password']").first
    expect(password_input).to_be_visible(timeout=15000)
    password_input.fill(password)
    page.locator("#passwordNext, button:has-text('Next')").first.click()



from __future__ import annotations

import os
from dataclasses import dataclass
from typing import Any

import boto3
from botocore.exceptions import ClientError

from bloom_lims.config import get_settings


DEFAULT_EMAIL = "johnm+test@lsmc.com"
DEFAULT_PASSWORD = "CodexPlaywright1!"
DEFAULT_NAME = "John M Test"
SERVICE_GROUPS = ("platform-admin",)


@dataclass(frozen=True)
class E2ECredentials:
    email: str
    password: str
    user_pool_id: str
    region: str
    sub: str


def _cognito_client(region: str, *, profile: str = ""):
    session = boto3.Session(
        profile_name=(os.getenv("E2E_AWS_PROFILE") or profile or os.getenv("AWS_PROFILE") or None),
        region_name=region,
    )
    return session.client("cognito-idp", region_name=region)


def _attributes(payload: dict[str, Any]) -> dict[str, str]:
    return {
        str(item.get("Name") or "").strip(): str(item.get("Value") or "").strip()
        for item in payload.get("UserAttributes", []) or []
        if str(item.get("Name") or "").strip()
    }


def _get_user(client, *, pool_id: str, email: str) -> dict[str, Any]:
    return client.admin_get_user(UserPoolId=pool_id, Username=email)


def _ensure_group(client, *, pool_id: str, group_name: str) -> None:
    try:
        client.get_group(UserPoolId=pool_id, GroupName=group_name)
    except ClientError as exc:
        if exc.response["Error"]["Code"] != "ResourceNotFoundException":
            raise
        client.create_group(
            UserPoolId=pool_id,
            GroupName=group_name,
            Description=f"E2E auto-created group {group_name}",
        )


def _ensure_membership(client, *, pool_id: str, email: str, group_name: str) -> None:
    try:
        client.admin_add_user_to_group(
            UserPoolId=pool_id,
            Username=email,
            GroupName=group_name,
        )
    except ClientError as exc:
        if exc.response["Error"]["Code"] != "ResourceNotFoundException":
            raise
        _ensure_group(client, pool_id=pool_id, group_name=group_name)
        client.admin_add_user_to_group(
            UserPoolId=pool_id,
            Username=email,
            GroupName=group_name,
        )


def ensure_test_user() -> E2ECredentials:
    settings = get_settings()
    region = str(os.getenv("E2E_COGNITO_REGION") or settings.auth.cognito_region or "").strip()
    pool_id = str(os.getenv("E2E_COGNITO_USER_POOL_ID") or settings.auth.cognito_user_pool_id or "").strip()
    profile = str(settings.aws.profile or "").strip()
    if not region or not pool_id:
        raise RuntimeError("Bloom E2E tests require Cognito region and user pool ID.")

    email = os.getenv("E2E_USER_EMAIL", DEFAULT_EMAIL).strip().lower()
    password = os.getenv("E2E_USER_PASSWORD", DEFAULT_PASSWORD).strip() or DEFAULT_PASSWORD
    client = _cognito_client(region, profile=profile)

    try:
        user_payload = _get_user(client, pool_id=pool_id, email=email)
    except ClientError as exc:
        if exc.response["Error"]["Code"] != "UserNotFoundException":
            raise
        client.admin_create_user(
            UserPoolId=pool_id,
            Username=email,
            TemporaryPassword=password,
            MessageAction="SUPPRESS",
            UserAttributes=[
                {"Name": "email", "Value": email},
                {"Name": "email_verified", "Value": "true"},
                {"Name": "name", "Value": DEFAULT_NAME},
            ],
        )
        user_payload = _get_user(client, pool_id=pool_id, email=email)

    client.admin_update_user_attributes(
        UserPoolId=pool_id,
        Username=email,
        UserAttributes=[
            {"Name": "email", "Value": email},
            {"Name": "email_verified", "Value": "true"},
            {"Name": "name", "Value": DEFAULT_NAME},
        ],
    )
    client.admin_set_user_password(
        UserPoolId=pool_id,
        Username=email,
        Password=password,
        Permanent=True,
    )

    for group_name in SERVICE_GROUPS:
        _ensure_membership(client, pool_id=pool_id, email=email, group_name=group_name)

    user_payload = _get_user(client, pool_id=pool_id, email=email)
    sub = _attributes(user_payload).get("sub", "")
    os.environ["E2E_USER_EMAIL"] = email
    os.environ["E2E_USER_PASSWORD"] = password
    return E2ECredentials(
        email=email,
        password=password,
        user_pool_id=pool_id,
        region=region,
        sub=sub,
    )



from __future__ import annotations

from types import SimpleNamespace

from bloom_lims.gui.jinja import refresh_template_globals, templates


def test_refresh_template_globals_exposes_deployment_banner(monkeypatch) -> None:
    fake_settings = SimpleNamespace(
        app_name="BLOOM LIMS",
        api=SimpleNamespace(version="1.2.3"),
        ui=SimpleNamespace(
            support_email="support@example.com",
            github_repo_url="https://github.com/Daylily-Informatics/bloom",
        ),
        deployment=SimpleNamespace(
            name="staging",
            color="#123456",
            is_production=False,
        ),
    )
    monkeypatch.setattr("bloom_lims.config.get_settings", lambda: fake_settings)

    refresh_template_globals()
    rendered = templates.get_template("modern/base.html").render(
        request=SimpleNamespace(url=SimpleNamespace(path="/")),
        user=None,
        udat=None,
        version="1.2.3",
    )

    assert "STAGING" in rendered
    assert "#123456" in rendered
    assert "/favicon.ico" in rendered



"""External integration commands for the Bloom CLI."""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from cli_core_yo.registry import CommandRegistry
    from cli_core_yo.spec import CliSpec

import typer
from rich.console import Console
from rich.table import Table

from bloom_lims.config import get_settings

integrations_app = typer.Typer(
    help="External integration commands.", no_args_is_help=True
)
atlas_app = typer.Typer(
    help="Atlas integration configuration and diagnostics.", no_args_is_help=True
)
console = Console()


@atlas_app.command("show")
def show_atlas_config() -> None:
    """Display the effective Atlas integration configuration."""
    settings = get_settings()
    table = Table(title="Bloom -> Atlas Integration")
    table.add_column("Field", style="cyan")
    table.add_column("Value")
    table.add_row("base_url", settings.atlas.base_url or "")
    table.add_row("organization_id", settings.atlas.organization_id or "")
    table.add_row("events_path", settings.atlas.events_path)
    table.add_row("timeout_seconds", str(settings.atlas.timeout_seconds))
    table.add_row("cache_ttl_seconds", str(settings.atlas.cache_ttl_seconds))
    table.add_row("verify_ssl", str(settings.atlas.verify_ssl))
    console.print(table)


@atlas_app.command("doctor")
def doctor_atlas_integration() -> None:
    """Report whether Atlas integration settings are usable."""
    settings = get_settings()
    missing: list[str] = []
    if not str(settings.atlas.base_url or "").strip():
        missing.append("atlas.base_url")
    if not str(settings.atlas.token or "").strip():
        missing.append("atlas.token")
    if missing:
        console.print(
            "[yellow]Atlas integration is partially configured.[/yellow] Missing: "
            + ", ".join(missing)
        )
        raise typer.Exit(1)
    console.print("[green]✓[/green] Atlas integration config is present")


integrations_app.add_typer(atlas_app, name="atlas")


def register(registry: CommandRegistry, spec: CliSpec) -> None:
    """cli-core-yo plugin: register the integrations command group."""
    registry.add_typer_app(
        None, integrations_app, "integrations", "External integration commands."
    )



"""Shared pytest runtime/bootstrap helpers for BLOOM integration suites."""

from __future__ import annotations

import os
import secrets
import socket
import tempfile
from pathlib import Path

from bloom_lims.config import (
    DEFAULT_BLOOM_TAPDB_LOCAL_PG_PORT,
    DEFAULT_BLOOM_WEB_PORT,
)

INTEGRATION_TEST_FILES = frozenset(
    {
        "test_api_v1.py",
        "test_gui_endpoints.py",
        "test_route_coverage_gaps_api.py",
        "test_route_coverage_gaps_gui.py",
    }
)

TEST_COGNITO_ENV_DEFAULTS = {
    "BLOOM_AUTH__COGNITO_USER_POOL_ID": "us-west-2_test-pool",
    "BLOOM_AUTH__COGNITO_CLIENT_ID": "bloom-test-client",
    "BLOOM_AUTH__COGNITO_DOMAIN": "bloom-test.auth.us-west-2.amazoncognito.com",
    "BLOOM_AUTH__COGNITO_REGION": "us-west-2",
    "BLOOM_AUTH__COGNITO_REDIRECT_URI": f"https://localhost:{DEFAULT_BLOOM_WEB_PORT}/auth/callback",
    "BLOOM_AUTH__COGNITO_LOGOUT_REDIRECT_URI": f"https://localhost:{DEFAULT_BLOOM_WEB_PORT}/",
}


def tapdb_config_path_needs_bootstrap(path_value: str | None) -> bool:
    """Return True when the configured Bloom TapDB config path should be replaced for tests."""
    if not str(path_value or "").strip():
        return True
    return not Path(path_value).expanduser().is_file()


def create_temp_tapdb_config(
    *, local_port: str | None = None, user: str | None = None
) -> Path:
    """Create a deterministic local TapDB namespaced config for tests."""
    resolved_port = str(
        local_port
        or os.environ.get("BLOOM_TAPDB_LOCAL_PG_PORT")
        or DEFAULT_BLOOM_TAPDB_LOCAL_PG_PORT
    ).strip()
    resolved_user = str(user or os.environ.get("USER") or "postgres").strip()
    tmp_path = (
        Path(tempfile.gettempdir()) / f"bloom_tapdb_config_{secrets.token_hex(16)}.yaml"
    )
    tmp_path.write_text(
        "\n".join(
            [
                "meta:",
                "  config_version: 3",
                "  client_id: bloom",
                "  database_name: bloom",
                "  euid_client_code: B",
                "environments:",
                "  dev:",
                "    engine_type: local",
                "    host: localhost",
                f'    port: "{resolved_port}"',
                f'    ui_port: "{DEFAULT_BLOOM_WEB_PORT}"',
                f'    user: "{resolved_user}"',
                '    password: ""',
                '    database: "tapdb_bloom_dev"',
                '    cognito_user_pool_id: "us-west-2_test-pool"',
                '    audit_log_euid_prefix: "BGX"',
                '    support_email: "support@dyly.bio"',
                "  test:",
                "    engine_type: local",
                "    host: localhost",
                f'    port: "{resolved_port}"',
                f'    ui_port: "{DEFAULT_BLOOM_WEB_PORT}"',
                f'    user: "{resolved_user}"',
                '    password: ""',
                '    database: "tapdb_bloom_test"',
                '    cognito_user_pool_id: "us-west-2_test-pool"',
                '    audit_log_euid_prefix: "BGX"',
                '    support_email: "support@dyly.bio"',
                "",
            ]
        ),
        encoding="utf-8",
    )
    os.chmod(tmp_path, 0o600)
    return tmp_path


def ensure_test_runtime_environment() -> Path:
    """Prepare deterministic Bloom config needed for strict app startup in tests."""
    os.environ.setdefault("MERIDIAN_ENVIRONMENT", "production")
    os.environ.setdefault("MERIDIAN_SANDBOX_PREFIX", "")
    os.environ.setdefault("BLOOM_TAPDB__CLIENT_ID", "bloom")
    os.environ.setdefault("BLOOM_TAPDB__DATABASE_NAME", "bloom")
    os.environ.setdefault("BLOOM_TAPDB__STRICT_NAMESPACE", "1")
    os.environ.setdefault(
        "BLOOM_TAPDB__LOCAL_PG_PORT", str(DEFAULT_BLOOM_TAPDB_LOCAL_PG_PORT)
    )

    for key, value in TEST_COGNITO_ENV_DEFAULTS.items():
        os.environ.setdefault(key, value)

    current_path = os.environ.get("BLOOM_TAPDB__CONFIG_PATH")
    if tapdb_config_path_needs_bootstrap(current_path):
        config_path = create_temp_tapdb_config()
        os.environ["BLOOM_TAPDB__CONFIG_PATH"] = str(config_path)
    else:
        config_path = Path(str(current_path)).expanduser()

    os.environ.setdefault("ECHO_SQL", "False")
    os.environ["BLOOM_DISABLE_RATE_LIMITING"] = "1"
    os.environ.setdefault("BLOOM_RATE_LIMIT", "no")

    from bloom_lims.config import get_settings

    get_settings.cache_clear()
    return config_path


def selected_items_need_local_tapdb(items: list[object]) -> bool:
    """Return True when the selected pytest items include DB-backed API/GUI suites."""
    for item in items:
        path = getattr(item, "fspath", None)
        if path is None:
            continue
        if Path(str(path)).name in INTEGRATION_TEST_FILES:
            return True
    return False


def tapdb_local_available(*, env_name: str = "dev") -> bool:
    """Return True when the configured local TapDB Postgres endpoint accepts TCP connections."""
    ensure_test_runtime_environment()

    from bloom_lims.config import get_tapdb_db_config

    config = get_tapdb_db_config(env_name=env_name)
    host = str(config.get("host") or "localhost").strip()
    port = int(
        config.get("port")
        or os.environ.get("BLOOM_TAPDB__LOCAL_PG_PORT")
        or DEFAULT_BLOOM_TAPDB_LOCAL_PG_PORT
    )

    try:
        with socket.create_connection((host, port), timeout=1):
            return True
    except OSError:
        return False


def ensure_local_tapdb_ready(*, env_name: str = "dev") -> bool:
    """Bootstrap local TapDB via the Bloom CLI when DB-backed integration suites need it."""
    ensure_test_runtime_environment()
    if tapdb_local_available(env_name=env_name):
        return False

    from bloom_lims.cli import db as db_commands

    try:
        db_commands.db_init(force=False)
    except SystemExit as exc:
        raise RuntimeError(f"`bloom db init` exited with status {exc.code}") from exc
    except Exception as exc:
        raise RuntimeError(f"`bloom db init` failed: {exc}") from exc

    if not tapdb_local_available(env_name=env_name):
        raise RuntimeError(
            "Local TapDB is still unavailable after `bloom db init`. "
            "Fix the runtime and retry `source ./activate <deploy-name> && bloom db init`."
        )

    return True



from __future__ import annotations

import os
import secrets
from typing import Any
from urllib.parse import urlparse

from fastapi import FastAPI, Request
from daylily_cognito import (
    CognitoWebSessionConfig,
    SessionPrincipal,
    clear_session_principal,
    configure_session_middleware,
    load_session_principal,
    store_session_principal,
)

from auth.cognito.client import CognitoAuth

from bloom_lims.config import get_settings

_SERVER_INSTANCE_ID = secrets.token_urlsafe(16)
_SESSION_SECRET_FALLBACK = secrets.token_urlsafe(32)


def get_server_instance_id() -> str:
    return _SERVER_INSTANCE_ID


def setup_bloom_session_middleware(app: FastAPI) -> None:
    configure_session_middleware(app, build_bloom_web_session_config())


def build_bloom_web_session_config(request: Request | None = None) -> CognitoWebSessionConfig:
    settings = get_settings()
    cognito = _resolve_cognito_auth(request)
    public_base_url = _origin(cognito.config.redirect_uri or "https://localhost:8912")
    explicit_secret = str(os.environ.get("BLOOM_SESSION_SECRET") or "").strip()
    base_secret = (
        explicit_secret
        or settings.auth.jwt_secret
        or settings.auth.cognito_client_secret
        or _SESSION_SECRET_FALLBACK
    )

    return CognitoWebSessionConfig(
        domain=cognito.config.domain,
        client_id=cognito.config.client_id,
        redirect_uri=cognito.config.redirect_uri,
        logout_uri=cognito.config.logout_redirect_uri or f"{public_base_url}/auth/logout",
        public_base_url=public_base_url,
        session_secret_key=base_secret,
        session_cookie_name="bloom_session",
        session_max_age=max(int(settings.auth.session_timeout_minutes or 30) * 60, 300),
        client_secret=cognito.config.client_secret or None,
        scope=" ".join(cognito.config.scopes or ["openid", "email", "profile"]),
        allow_insecure_http=public_base_url.startswith("http://"),
        server_instance_id=get_server_instance_id(),
    )


def store_bloom_session(
    request: Request,
    principal: SessionPrincipal,
    *,
    user_data: dict[str, Any],
) -> SessionPrincipal:
    config = build_bloom_web_session_config(request)
    stored = store_session_principal(request, config, principal)
    request.session["user_data"] = _normalize_user_data(dict(user_data), stored)
    return stored


def load_bloom_user_data(request: Request) -> dict[str, Any] | None:
    try:
        principal = load_session_principal(request)
    except RuntimeError:
        principal = None
    if principal is None:
        request.session.pop("user_data", None)
        return None

    user_data = request.session.get("user_data")
    if isinstance(user_data, dict):
        normalized = _normalize_user_data(dict(user_data), principal)
        request.session["user_data"] = normalized
        return normalized

    normalized = _user_data_from_principal(principal)
    request.session["user_data"] = normalized
    return normalized


def clear_bloom_session(request: Request) -> None:
    clear_session_principal(request)
    request.session.pop("_cognito_oauth_state", None)
    request.session.pop("_cognito_post_auth_redirect", None)
    request.session.pop("user_data", None)
    request.session.pop("bloom_post_auth_redirect", None)


def _resolve_cognito_auth(request: Request | None = None) -> CognitoAuth:
    if request is not None:
        from bloom_lims.gui.deps import _get_request_cognito_auth

        return _get_request_cognito_auth(request)

    settings = get_settings()
    return CognitoAuth.from_settings(settings.auth)


def _user_data_from_principal(principal: SessionPrincipal) -> dict[str, Any]:
    app_context = principal.app_context if isinstance(principal.app_context, dict) else {}
    user_data = app_context.get("user_data")
    if not isinstance(user_data, dict):
        user_data = {}
    return _normalize_user_data(dict(user_data), principal)


def _normalize_user_data(user_data: dict[str, Any], principal: SessionPrincipal) -> dict[str, Any]:
    normalized = dict(user_data)
    normalized.pop("access_token", None)
    normalized.pop("id_token", None)
    normalized.pop("refresh_token", None)
    normalized["email"] = str(normalized.get("email") or principal.email)
    normalized["name"] = normalized.get("name") or principal.name or normalized["email"]
    normalized["sub"] = str(normalized.get("sub") or principal.user_sub)
    normalized["cognito_sub"] = str(normalized.get("cognito_sub") or principal.user_sub)
    normalized["user_id"] = str(normalized.get("user_id") or normalized["sub"])
    normalized["roles"] = list(normalized.get("roles") or principal.roles or [])
    if normalized["roles"]:
        normalized["role"] = str(normalized.get("role") or normalized["roles"][0]).upper()
        normalized["roles"] = [str(role).upper() for role in normalized["roles"] if str(role).strip()]
    else:
        normalized["role"] = str(normalized.get("role") or "READ_WRITE").upper()
        normalized["roles"] = [normalized["role"]]
    normalized["groups"] = list(normalized.get("groups") or principal.cognito_groups or [])
    return normalized


def _origin(url: str) -> str:
    parsed = urlparse((url or "").strip())
    scheme = parsed.scheme or "https"
    netloc = parsed.netloc or "localhost:8912"
    return f"{scheme}://{netloc}"



"""Testing commands for BLOOM CLI."""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from cli_core_yo.registry import CommandRegistry
    from cli_core_yo.spec import CliSpec

import subprocess
import sys
from pathlib import Path

import typer

test_app = typer.Typer(help="Test execution commands", no_args_is_help=True)

PROJECT_ROOT = Path(__file__).resolve().parents[2]


@test_app.command(
    "run",
    context_settings={"allow_extra_args": True, "ignore_unknown_options": True},
)
def run(ctx: typer.Context) -> None:
    """Run Bloom test suites via pytest."""
    args = list(ctx.args) or ["tests"]
    raise typer.Exit(
        subprocess.run(
            [sys.executable, "-m", "pytest", *args],
            cwd=PROJECT_ROOT,
            check=False,
        ).returncode
    )


def register(registry: CommandRegistry, spec: CliSpec) -> None:
    """cli-core-yo plugin: register the test command group."""
    registry.add_typer_app(None, test_app, "test", "Test execution commands.")



from __future__ import annotations

from bloom_lims.auth.services.groups import GroupService


def test_resolve_user_roles_and_groups_maps_uppercase_system_group_codes(monkeypatch):
    service = GroupService(db=None)  # type: ignore[arg-type]
    monkeypatch.setattr(
        GroupService,
        "get_group_codes_for_user",
        lambda self, user_id: ["BLOOM-ADMIN"],
    )

    resolution = service.resolve_user_roles_and_groups(
        user_id="johnm@lsmc.com",
        fallback_role=None,
    )

    assert resolution.roles == ["ADMIN"]
    assert resolution.groups == ["BLOOM-ADMIN"]



"""Additional coverage for Bloom DB CLI command wiring."""

from __future__ import annotations

import importlib
from pathlib import Path
from types import SimpleNamespace

import pytest
from typer.testing import CliRunner

from bloom_lims.cli import build_app
from bloom_lims.config import DEFAULT_BLOOM_TAPDB_LOCAL_PG_PORT, DEFAULT_BLOOM_WEB_PORT

db_commands = importlib.import_module("bloom_lims.cli.db")


@pytest.fixture
def runner() -> CliRunner:
    return CliRunner()


@pytest.fixture
def cli_app(monkeypatch: pytest.MonkeyPatch, tmp_path: Path):
    monkeypatch.setenv("HOME", str(tmp_path))
    return build_app()


def test_run_tapdb_raises_for_nonzero_check(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(db_commands, "_tapdb_base_cmd", lambda: ["tapdb"])
    monkeypatch.setattr(db_commands, "_runtime_env", lambda: {})
    monkeypatch.setattr(
        db_commands,
        "apply_runtime_environment",
        lambda _settings: SimpleNamespace(
            config_path="/tmp/bloom-tapdb.yaml",
            env="dev",
        ),
    )
    monkeypatch.setattr(
        db_commands.subprocess,
        "run",
        lambda *_args, **_kwargs: SimpleNamespace(returncode=7),
    )

    with pytest.raises(SystemExit) as exc:
        db_commands._run_tapdb(["db", "setup"], check=True)

    assert exc.value.code == 7


def test_run_tapdb_returns_nonzero_when_check_false(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(db_commands, "_tapdb_base_cmd", lambda: ["tapdb"])
    monkeypatch.setattr(db_commands, "_runtime_env", lambda: {})
    monkeypatch.setattr(
        db_commands,
        "apply_runtime_environment",
        lambda _settings: SimpleNamespace(
            config_path="/tmp/bloom-tapdb.yaml",
            env="dev",
        ),
    )
    monkeypatch.setattr(
        db_commands.subprocess,
        "run",
        lambda *_args, **_kwargs: SimpleNamespace(returncode=3),
    )

    assert db_commands._run_tapdb(["db", "setup"], check=False) == 3


def test_update_tapdb_namespace_config_uses_config_update(monkeypatch: pytest.MonkeyPatch) -> None:
    calls: list[tuple[list[str], bool]] = []
    monkeypatch.setattr(db_commands, "_tapdb_audit_log_euid_prefix", lambda _env: "audit.bloom")
    monkeypatch.setattr(db_commands, "_tapdb_support_email", lambda _env: "support@example.com")
    monkeypatch.setattr(
        db_commands,
        "_run_tapdb",
        lambda args, check=True: calls.append((args, check)) or 0,
    )

    db_commands._update_tapdb_namespace_config("dev")

    assert calls == [
        (
            [
                "config",
                "update",
                "--env",
                "dev",
                "--audit-log-euid-prefix",
                "audit.bloom",
                "--support-email",
                "support@example.com",
            ],
            True,
        )
    ]


def test_ensure_tapdb_namespace_config_initializes_then_updates(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    calls: list[tuple[list[str], bool]] = []
    monkeypatch.setattr(
        db_commands,
        "apply_runtime_environment",
        lambda _settings: SimpleNamespace(
            client_id="bloom",
            database_name="bloom",
            config_path="/tmp/.config/tapdb/bloom/bloom-local2/tapdb-config.yaml",
            env="dev",
            aws_profile="lsmc",
            aws_region="us-west-2",
        ),
    )
    monkeypatch.setattr(db_commands, "_local_pg_port", lambda _env: str(DEFAULT_BLOOM_TAPDB_LOCAL_PG_PORT))
    monkeypatch.setattr(db_commands, "_local_ui_port", lambda _env: str(DEFAULT_BLOOM_WEB_PORT))
    monkeypatch.setattr(db_commands, "_tapdb_audit_log_euid_prefix", lambda _env: "BBL")
    monkeypatch.setattr(db_commands, "_tapdb_support_email", lambda _env: "support@example.com")
    monkeypatch.setattr(
        db_commands,
        "_run_tapdb",
        lambda args, check=True: calls.append((args, check)) or 0,
    )

    db_commands._ensure_tapdb_namespace_config("dev")

    assert calls == [
        (
            [
                "config",
                "init",
                "--env",
                "dev",
                "--db-port",
                "dev=5566",
                "--ui-port",
                "dev=8912",
            ],
            True,
        ),
        (
            [
                "config",
                "update",
                "--env",
                "dev",
                "--audit-log-euid-prefix",
                "BBL",
                "--support-email",
                "support@example.com",
            ],
            True,
        ),
    ]


def test_ensure_tapdb_namespace_config_creates_scoped_parent(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    config_path = tmp_path / ".config" / "tapdb" / "bloom" / "bloom-local2" / "tapdb-config.yaml"
    monkeypatch.setattr(
        db_commands,
        "apply_runtime_environment",
        lambda _settings: SimpleNamespace(
            client_id="bloom",
            database_name="bloom",
            config_path=str(config_path),
            env="dev",
            aws_profile="lsmc",
            aws_region="us-west-2",
        ),
    )
    monkeypatch.setattr(db_commands, "_local_pg_port", lambda _env: str(DEFAULT_BLOOM_TAPDB_LOCAL_PG_PORT))
    monkeypatch.setattr(db_commands, "_local_ui_port", lambda _env: str(DEFAULT_BLOOM_WEB_PORT))
    monkeypatch.setattr(db_commands, "_tapdb_audit_log_euid_prefix", lambda _env: "BBL")
    monkeypatch.setattr(db_commands, "_tapdb_support_email", lambda _env: "support@example.com")
    monkeypatch.setattr(db_commands, "_run_tapdb", lambda *_args, **_kwargs: 0)

    db_commands._ensure_tapdb_namespace_config("dev")

    assert config_path.parent.exists()


def test_resolve_tapdb_schema_source_skips_dayhoff_artifacts(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
) -> None:
    sibling_schema = tmp_path / "daylily-tapdb" / "schema" / "tapdb_schema.sql"
    sibling_schema.parent.mkdir(parents=True, exist_ok=True)
    sibling_schema.write_text("-- sibling schema\n", encoding="utf-8")

    artifact_pkg = tmp_path / ".dayhoff" / "local" / "lsmc5" / "repos" / "daylily-tapdb" / "daylily_tapdb" / "__init__.py"
    artifact_pkg.parent.mkdir(parents=True, exist_ok=True)
    artifact_pkg.write_text("# artifact package\n", encoding="utf-8")

    monkeypatch.setattr(db_commands, "_bloom_root", lambda: tmp_path / "bloom")
    monkeypatch.setattr(
        db_commands.importlib,
        "import_module",
        lambda _name: SimpleNamespace(__file__=str(artifact_pkg)),
    )

    assert db_commands._resolve_tapdb_schema_source() == sibling_schema


def test_db_seed_calls_tapdb_template_loader(
    runner: CliRunner,
    cli_app,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    tapdb_seed: list[tuple[str, bool, bool]] = []
    monkeypatch.setattr(db_commands, "_current_env", lambda: "dev")
    monkeypatch.setattr(
        db_commands,
        "_seed_tapdb_templates",
        lambda env_name, include_workflow, overwrite: tapdb_seed.append(
            (env_name, include_workflow, overwrite)
        ),
    )

    result = runner.invoke(cli_app, ["db", "seed"])
    assert result.exit_code == 0
    assert tapdb_seed == [("dev", False, False)]


@pytest.mark.parametrize(
    ("force", "expected_args"),
    [
        (False, ["db", "schema", "reset", "dev"]),
        (True, ["db", "schema", "reset", "dev", "--force"]),
    ],
)
def test_db_nuke_calls_tapdb_schema_reset(
    runner: CliRunner,
    cli_app,
    monkeypatch: pytest.MonkeyPatch,
    force: bool,
    expected_args: list[str],
) -> None:
    calls: list[tuple[list[str], bool]] = []
    monkeypatch.setattr(db_commands, "_current_env", lambda: "dev")
    monkeypatch.setattr(
        db_commands,
        "_run_tapdb",
        lambda args, check=True: calls.append((args, check)) or 0,
    )

    argv = ["db", "nuke"]
    if force:
        argv.append("--force")

    result = runner.invoke(cli_app, argv)
    assert result.exit_code == 0
    assert calls == [(expected_args, True)]



"""Quality commands for BLOOM CLI."""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from cli_core_yo.registry import CommandRegistry
    from cli_core_yo.spec import CliSpec

import subprocess
import sys
from pathlib import Path

import typer

quality_app = typer.Typer(
    help="Code-quality and validation commands", no_args_is_help=True
)

PROJECT_ROOT = Path(__file__).resolve().parents[2]


def _run(cmd: list[str]) -> None:
    raise typer.Exit(
        subprocess.run(
            cmd,
            cwd=PROJECT_ROOT,
            check=False,
        ).returncode
    )


@quality_app.command("check")
def check() -> None:
    """Run the standard Bloom quality gates."""
    commands = (
        [sys.executable, "-m", "ruff", "check", "bloom_lims", "tests"],
        [sys.executable, "-m", "bandit", "-c", "pyproject.toml", "-r", "bloom_lims"],
    )
    for cmd in commands:
        result = subprocess.run(cmd, cwd=PROJECT_ROOT, check=False)
        if result.returncode != 0:
            raise typer.Exit(result.returncode)


@quality_app.command(
    "ruff",
    context_settings={"allow_extra_args": True, "ignore_unknown_options": True},
)
def ruff(ctx: typer.Context) -> None:
    """Run Ruff with optional extra arguments."""
    args = list(ctx.args) or ["check", "bloom_lims", "tests"]
    _run([sys.executable, "-m", "ruff", *args])


@quality_app.command(
    "bandit",
    context_settings={"allow_extra_args": True, "ignore_unknown_options": True},
)
def bandit(ctx: typer.Context) -> None:
    """Run Bandit with optional extra arguments."""
    args = list(ctx.args) or ["-c", "pyproject.toml", "-r", "bloom_lims"]
    _run([sys.executable, "-m", "bandit", *args])


def register(registry: CommandRegistry, spec: CliSpec) -> None:
    """cli-core-yo plugin: register the quality command group."""
    registry.add_typer_app(None, quality_app, "quality", "Code-quality commands.")



from __future__ import annotations

import pytest
from playwright.sync_api import Page, expect

from tests.e2e.auth_helpers import assert_authenticated_page, perform_login, perform_logout


@pytest.mark.e2e
def test_login_round_trip(page: Page, base_url: str, e2e_email: str, e2e_password: str):
    perform_login(page, base_url=base_url, email=e2e_email, password=e2e_password)
    expect(page).to_have_url(f"{base_url}/")
    assert_authenticated_page(page)


@pytest.mark.e2e
def test_logout_round_trip(page: Page, base_url: str, e2e_email: str, e2e_password: str):
    perform_login(page, base_url=base_url, email=e2e_email, password=e2e_password)
    perform_logout(page, base_url=base_url)
    current_url = page.url
    assert (
        current_url.startswith(f"{base_url}/login")
        or "auth.us-west-2.amazoncognito.com/login" in current_url
        or "auth.us-west-2.amazoncognito.com/logout" in current_url
    ), f"Unexpected logout landing URL: {current_url}"

    response = page.goto(f"{base_url}/")
    assert response is not None
    assert response.status in {200, 302, 303}
    page.wait_for_load_state("domcontentloaded")
    current_url = page.url
    assert (
        current_url.startswith(f"{base_url}/login")
        or "auth.us-west-2.amazoncognito.com/login" in current_url
        or "auth.us-west-2.amazoncognito.com/logout" in current_url
    ), f"Unexpected protected-route logout URL: {current_url}"
